import numpy as np
# from gpaw.poisson import FastPoissonSolver, FDPoissonSolver, create_poisson_solver, BasePoissonSolver
# from gpaw.grid_descriptor import GridDescriptor


# class PoissonSolver(BasePoissonSolver):
#     def __init__(self, solver_name='fast', max_num_iter=10, **kwargs):
#         BasePoissonSolver.__init__(self, **kwargs)
#         self.max_num_iter = max_num_iter
#         self.solver_name = solver_name
#         self.solver = create_poisson_solver(name='fast', **kwargs)
#         self.delta = []
#
#     def set_grid_descriptor(self, gd):
#         self.gd = gd
#         self.solver.set_grid_descriptor(gd)
#
#     def solve(self, pot, dens0, permitivity):
#
#         self.delta = []
#
#         alpha = 0.01
#         dens0 = dens0 / permitivity
#         conv_criteria = np.max(dens0) * 0.0001
#         dens = dens0
#
#         dx = self.gd.h_cv[0][0]
#
#         for j in range(self.max_num_iter):
#             print(j)
#             self.solver.solve(pot, dens)
#
#             aaa_x, aaa_y, aaa_z = np.gradient(permitivity, dx)
#             bbb_x, bbb_y, bbb_z = np.gradient(pot, dx)
#
#             delta = (aaa_x * bbb_x + aaa_y * bbb_y + aaa_z * bbb_z) / permitivity
#
#             self.delta.append(np.max(delta))
#             dens = dens0 + delta * alpha
#
#             if j > 1:
#                 if np.abs(self.delta[-1] - self.delta[-2]) < conv_criteria:
#                     return
#
#         raise RuntimeError('PoissonSolver have not converged.')
#
#     def solve_mesh(self, X, Y, Z, pot, dens0, permitivity):
#
#         gd = GridDescriptor([X.shape[0] + 1, X.shape[1] + 1, X.shape[2] + 1],
#                             cell_cv=[X[0, -1, 0] - X[0, 0, 0],
#                                      Y[-1, 0, 0] - Y[0, 0, 0],
#                                      Z[0, 0, -1] - Z[0, 0, 0]], pbc_c=False)
#         self.set_grid_descriptor(gd)
#         self.solve(pot, dens0, permitivity)
#
#     def todict(self):
#         self.solver.todict()
#
#     def get_description(self):
#         return self.__class__.__name__
#
#     def estimate_memory(self, mem):
#         self.solver.estimate_memory(mem)


def laplacian(input_mat, dx):
    """

    Parameters
    ----------
    input_mat :
        
    dx :
        

    Returns
    -------

    """

    top = input_mat[0:-2, 1:-1, 1:-1]
    left = input_mat[1:-1, 0:-2, 1:-1]
    bottom = input_mat[2:, 1:-1, 1:-1]
    right = input_mat[1:-1, 2:, 1:-1]
    center = input_mat[1:-1, 1:-1, 1:-1]
    back = input_mat[1:-1, 1:-1, 0:-2]
    forward = input_mat[1:-1, 1:-1, 2:]
    return np.pad(top + left + bottom + right + back + forward - 6 * center,
                  ((1, 1), (1, 1), (1, 1)),
                  'constant',
                  constant_values=((0, 0), (0, 0), (0, 0))) / dx**2


def gf(x, y, z, x0, y0, z0, sigma):
    """

    Parameters
    ----------
    x :
        
    y :
        
    z :
        
    x0 :
        
    y0 :
        
    z0 :
        
    sigma :
        

    Returns
    -------

    """
    return np.exp(-((x-x0)**2+(y-y0)**2+(z-z0)**2) / (2*sigma**2))


if __name__=="__main__":

    num_points = 128
    # x = np.linspace(0, 100, num_points - 1)
    # dx = x[3] - x[2]
    # X, Y, Z = np.meshgrid(x, x, x)
    # gd = GridDescriptor([num_points, num_points, num_points], cell_cv=[100, 100, 100], pbc_c=False)
    #
    # dens0 = 10*gf(X, Y, Z, 50, 50, 50, 3)
    # eps = 50.8*(1.0 - 0.5*np.tanh(3000.0*(Y - 60))-0.5) + 3.5*(0.5*np.tanh(3000.0*(Y - 60))+0.5)
    # dens0 = dens0 / eps
    # pot = np.zeros(dens0.shape)
    #
    #
    # ps = PoissonSolver()
    # # ps.set_grid_descriptor(gd)
    # # ps.solve(pot, dens0, eps)
    # ps.solve_mesh(X, Y, Z, pot, dens0, eps)
    #
    # import matplotlib.pyplot as plt
    # plt.plot(np.array(delta))
    # plt.contour(pot[:, :, 64], 50)
    # plt.show()



"""
Module contains the class `Orbitals` that allows to generate any user defined basis set based on
the linear combination of atomic orbitals (LCAO). Also, the module contains a set of predefined basis sets
`SiliconSP3D5S`, `HydrogenS`, `Bismuth`.
"""
import sys
from nanonet.tb.aux_functions import print_table


class Orbitals(object):
    """This is the parent class for all basis sets for all atoms. It also contains a factory function,
    which generates objects of the class Orbitals from a list of labels and
    the dictionary `orbital_sets` making a correspondence between an atom and its basis set

    Parameters
    ----------

    Returns
    -------

    """

    orbital_sets = {}

    def __init__(self, title):
        self.title = title
        self.orbitals = []
        self.num_of_orbitals = 0
        Orbitals.orbital_sets[self.title] = self

    def add_orbital(self, title, energy=0.0, principal=0, orbital=0, magnetic=0, spin=0):
        """Adds an orbital to the set of orbitals

        Parameters
        ----------
        title :
            a string representing an orbital label,
            it usually specifies its symmetry, e.g. `s`, `px`, `py` etc.
        energy :
            energy of the orbital (Default value = 0.0)
        principal :
            principal quantum number `n-1` (Default value = 0)
        orbital :
            orbital quantum number `l` (Default value = 0)
        magnetic :
            magnetic quantum number `m` (Default value = 0)
        spin :
            spin quantum number `s` (Default value = 0)

        Returns
        -------

        """

        orbital = {'title': title,
                   'energy': energy,
                   'n': principal,
                   'l': orbital,
                   'm': magnetic,
                   's': spin}

        self.orbitals.append(orbital)
        self.num_of_orbitals += 1

        Orbitals.orbital_sets[self.title] = self

    def generate_info(self):
        """ """

        return print_table(self.orbitals)

    @staticmethod
    def atoms_factory(labels):
        """Taking a list of labels creates a dictionary of `Orbitals` objects
        from those labels. The set of orbitals for each atom and corresponding class is
        specified in the class variable `orbital_sets`

        Parameters
        ----------
        labels : list(str)
            list of labels

        Returns
        -------
        type
            dictionary of `Orbitals` objects

        """

        output = {}

        for label in labels:

            try:
                key = ''.join([i for i in label if not i.isdigit()])
                atom = Orbitals.orbital_sets[key]
                if not isinstance(atom, Orbitals):
                    raise KeyError
            except KeyError:
                # TODO: simplify these statements below
                if label.lower().startswith('si'):
                    atom = getattr(sys.modules[__name__], Orbitals.orbital_sets['Si'])()
                elif label.lower().startswith('h'):
                    atom = getattr(sys.modules[__name__], Orbitals.orbital_sets['H'])()
                elif label.lower().startswith('b'):
                    atom = getattr(sys.modules[__name__], Orbitals.orbital_sets['Bi'])()
                else:
                    raise ValueError("There is no library entry for the atom " + label)

            output[atom.title] = atom

        return output


class SiliconSP3D5S(Orbitals):
    """Class defines the `sp3d5s*` basis set for the silicon atom"""

    def __init__(self):

        super(SiliconSP3D5S, self).__init__("Si")

        self.add_orbital("s", energy=-2.0196, spin=0)
        self.add_orbital("c", energy=19.6748, principal=1, spin=0)
        self.add_orbital("px", energy=4.5448, orbital=1, magnetic=-1, spin=0)
        self.add_orbital("py", energy=4.5448, orbital=1, magnetic=1, spin=0)
        self.add_orbital("pz", energy=4.5448, orbital=1, magnetic=0, spin=0)
        self.add_orbital("dz2", energy=14.1836, orbital=2, magnetic=-1, spin=0)
        self.add_orbital("dxz", energy=14.1836, orbital=2, magnetic=-2, spin=0)
        self.add_orbital("dyz", energy=14.1836, orbital=2, magnetic=2, spin=0)
        self.add_orbital("dxy", energy=14.1836, orbital=2, magnetic=1, spin=0)
        self.add_orbital("dx2my2", energy=14.1836, orbital=2, magnetic=0, spin=0)


class HydrogenS(Orbitals):
    """Class defines the simplest basis set for the hydrogen atom,
    consisting of a single s-orbital

    Parameters
    ----------

    Returns
    -------

    """

    def __init__(self):

        super(HydrogenS, self).__init__("H")

        self.add_orbital("s", energy=0.9998)
        # self.add_orbital("s", energy=0.9998, spin=1)


class Bismuth(Orbitals):
    """Class defines the `sp3` basis set for the bismuth atom"""

    def __init__(self):

        super(Bismuth, self).__init__("Bi")

        self.add_orbital("s", energy=-10.906, principal=0, orbital=0, magnetic=0, spin=0)
        self.add_orbital("px", energy=-0.486, principal=0, orbital=1, magnetic=-1, spin=0)
        self.add_orbital("py", energy=-0.486, principal=0, orbital=1, magnetic=1, spin=0)
        self.add_orbital("pz", energy=-0.486, principal=0, orbital=1, magnetic=0, spin=0)
        self.add_orbital("s", energy=-10.906, principal=0, orbital=0, magnetic=0, spin=1)
        self.add_orbital("px", energy=-0.486, principal=0, orbital=1, magnetic=-1, spin=1)
        self.add_orbital("py", energy=-0.486, principal=0, orbital=1, magnetic=1, spin=1)
        self.add_orbital("pz", energy=-0.486, principal=0, orbital=1, magnetic=0, spin=1)



"""
The module contains classes defining
geometrical structure and boundary conditions for the tight-binding model.
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from collections import OrderedDict
import logging
import numpy as np
import scipy.spatial
from nanonet.tb.abstract_interfaces import AbstractStructureDesigner
from nanonet.tb.aux_functions import xyz2np, count_species, is_in_coords, print_dict


class StructDesignerXYZ(AbstractStructureDesigner):
    """The class builds an atomic structure from either
    the filename of a xyz-file or
    xyz data itself represented as a Python string.
    The class arrange atomic coordinates in kd-tree and
    sorts them if needed according to a specified sorting procedure.

    Parameters
    ----------

    Returns
    -------

    Attributes
    ----------
    nn_distance : float
        nearest neighbour search radius (default 0)
    num_of_species : int
        number of chemical elements corresponding to the number of distinct basis sets.
    atom_list : OrderedDict
        list of atomic species and their coordinates
    kd_tree : scipy.spatial.ckdtree.cKDTree
        kd-tree for fast nearest-neighbour search
    left_lead : list
        list of atomic indices connected to the left lead,
        needed for sorting atomic coordinates (default [])
    right_lead : list
        list of atomic indices connected to the right lead,
        needed for sorting atomic coordinates (default [])
    sort_func : func
        function for sorting atomic coordinates (default None)
    """

    def __init__(self, **kwargs):
        # ------------ parse xyz file or string --------------
        xyz = kwargs.get('xyz', None)

        try:
            with open(xyz, 'r') as read_file:
                reader = read_file.read()
        except IOError:
            reader = xyz

        labels, coords = xyz2np(reader)

        num_lines = reader.count('\n')
        if num_lines > 11:
            logging.info("The xyz-file:\n {}".format('\n'.join(reader.split('\n')[:11])))
            logging.info("                  .                    ")
            logging.info("                  .                    ")
            logging.info("                  .                    ")
            logging.info("There are {} more coordinates".format(str(num_lines-10)))
        else:
            logging.info("The xyz-file:\n {}".format(reader))
        logging.info("---------------------------------\n")

        # ------------- count species and nodes --------------
        self._nn_distance = kwargs.get('nn_distance', 0)           # maximal distance to a neighbor
        self._num_of_species = count_species(labels)               # dictionary of elements and
                                                                   # their number per unit cell
        self._num_of_nodes = sum(self.num_of_species.values())
        # ------- make list of coordinates and kd-tree -------
        self._atom_list = OrderedDict(list(zip(labels, coords)))
        self._kd_tree = scipy.spatial.cKDTree(np.array(list(self._atom_list.values())),
                                              leafsize=1,
                                              balanced_tree=True)

        # ------- rearrange list of atomic coordinates -------
        self.left_lead = kwargs.get('left_lead', [])
        self.right_lead = kwargs.get('right_lead', [])
        self.sort_func = kwargs.get('sort_func', None)
        self.reorder = None

        if self.sort_func is not None:
            self._sort(labels, coords)

    def _sort(self, labels, coords):
        """

        Parameters
        ----------
        labels :
            
        coords :
            

        Returns
        -------

        """

        coords = np.array(coords)
        h_matrix = np.zeros((coords.shape[0], coords.shape[0]))

        self._nn_distance = 2 * self._nn_distance
        for j in range(len(coords)):
            ans = self.get_neighbours(j)
            h_matrix[j, ans] = 1

        self._nn_distance = self._nn_distance / 2

        indices = self.sort_func(coords=coords,
                                 left_lead=self.left_lead,
                                 right_lead=self.right_lead,
                                 mat=h_matrix)

        self.reorder = indices
        if (isinstance(self.left_lead, list) or isinstance(self.left_lead, np.ndarray)) and len(self.left_lead) > 0:
            self.left_lead = np.squeeze(np.concatenate([np.where(indices == item) for item in self.left_lead]))
        if (isinstance(self.right_lead, list) or isinstance(self.right_lead, np.ndarray)) and len(self.right_lead) > 0:
            self.right_lead = np.squeeze(np.concatenate([np.where(indices == item) for item in self.right_lead]))
        coords = coords[indices]
        labels = [labels[i] for i in indices]

        self._atom_list = OrderedDict(list(zip(labels, coords)))
        self._kd_tree = scipy.spatial.cKDTree(np.array(list(self._atom_list.values())), leafsize=1, balanced_tree=True)

    def add_leads(self, left_lead, right_lead):
        """

        Parameters
        ----------
        left_lead :
            
        right_lead :
            

        Returns
        -------

        """
        self.left_lead = left_lead
        self.right_lead = right_lead

    @property
    def atom_list(self):
        """ """
        return self._atom_list

    @property
    def num_of_nodes(self):
        """ """
        return self._num_of_nodes

    @property
    def num_of_species(self):
        """ """
        return self._num_of_species

    def get_neighbours(self, query):
        """

        Parameters
        ----------
        query :
            

        Returns
        -------

        """

        ans = self._get_neighbours(query)

        if ans[0][0] == np.inf:
            ans1 = []
        else:
            ans1 = [ans[1][0]]

        for item in zip(ans[0], ans[1]):
            if self._nn_distance * 0.1 < item[0] < self._nn_distance:
                ans1.append(item[1])

        return ans1


class CyclicTopology(AbstractStructureDesigner):
    """The class provides functionality for determining
    the periodic boundary conditions for a crystal cell.
    The object of the class is instantiated by
    a set of the primitive cell vectors.

    Parameters
    ----------

    Returns
    -------

    """

    def __init__(self, primitive_cell_vectors, labels, coords, nn_distance):

        self._nn_distance = nn_distance
        self.pcv = primitive_cell_vectors

        # compute vectors' lengths
        self.sizes = []
        for item in self.pcv:
            self.sizes.append(np.linalg.norm(item))

        self.interfacial_atoms_ind = []
        self.virtual_and_interfacial_atoms = OrderedDict()

        self.shift = np.zeros(3)

        self._generate_atom_list(labels, coords)

        self._kd_tree = scipy.spatial.cKDTree(list(self.virtual_and_interfacial_atoms.values()),
                                              leafsize=100, balanced_tree=True)

        logging.info("Primitive_cell_vectors: \n {} \n".format(primitive_cell_vectors))
        logging.debug("Virtual and interfacial atoms: \n "
                      "{} ".format(print_dict(self.virtual_and_interfacial_atoms)))
        logging.info("---------------------------------\n")

    @property
    def atom_list(self):
        """ """
        return self.virtual_and_interfacial_atoms

    def _generate_atom_list(self, labels, coords):
        """

        Parameters
        ----------
        labels :
            labels of atoms
        coords :
            coordinates of atoms

        Returns
        -------

        """

        # matrices of distances between atoms and interfaces
        distances1 = np.empty((len(coords), len(self.pcv)), dtype=float)
        distances2 = np.empty((len(coords), len(self.pcv)), dtype=float)

        for j1, coord in enumerate(coords):    # for each atom in the unit cell
            for j2, basis_vec in enumerate(self.pcv):    # for lattice basis vector

                # compute distance to the primary plane of the unit cell
                distances1[j1, j2] = np.inner(coord - self.shift, basis_vec) / self.sizes[j2]
                # compute distance to the adjacent plane of the  unit cell
                distances2[j1, j2] = np.inner(coord - self.shift - basis_vec, basis_vec) / self.sizes[j2]

        for item in distances1.T:
            self.shift += coords[np.argmin(item)]

        for j1, coord in enumerate(coords):    # for each atom in the unit cell
            for j2, basis_vec in enumerate(self.pcv):    # for lattice basis vector

                # compute distance to the primary plane of the unit cell
                distances1[j1, j2] = np.inner(coord - self.shift, basis_vec) / self.sizes[j2]
                # compute distance to the adjacent plane of the  unit cell
                distances2[j1, j2] = np.inner(coord - self.shift - basis_vec, basis_vec) / self.sizes[j2]

        # transform distance to the boolean variable defining whether atom belongs to the interface or not
        distances1 = np.abs(distances1 - np.min(distances1)) < self._nn_distance * 0.25
        distances2 = np.abs(np.abs(distances2) - np.min(np.abs(distances2))) < self._nn_distance * 0.25

        # distances1 = np.ones(distances1.shape)
        # distances2 = np.ones(distances1.shape)

        # form new lists of atoms
        count = 0
        for j, item in enumerate(coords):

            if any(distances1[j]):
                self.virtual_and_interfacial_atoms.update({str(j) + "_" + labels[j]: item})
                self.interfacial_atoms_ind.append(j)

                for surf in np.where(distances1[j])[0]:

                    count = self._translate_atom_1st_order(item,
                                                           np.array(self.pcv[surf]),
                                                           "_" + str(j) + "_" + labels[j],
                                                           coords,
                                                           count)

                    count = self._translate_atom_2d_order(item,
                                                          np.array(self.pcv[surf]),
                                                          "_" + str(j) + "_" + labels[j],
                                                          coords,
                                                          count)

            if any(distances2[j]):
                self.virtual_and_interfacial_atoms.update({str(j) + "_" + labels[j]: item})
                self.interfacial_atoms_ind.append(j)
                for surf in np.where(distances2[j])[0]:

                    count = self._translate_atom_1st_order(item,
                                                           -1 * np.array(self.pcv[surf]),
                                                           "_" + str(j) + "_" + labels[j],
                                                           coords,
                                                           count)

                    count = self._translate_atom_2d_order(item,
                                                          -1 * np.array(self.pcv[surf]),
                                                          "_" + str(j) + "_" + labels[j],
                                                          coords,
                                                          count)

        # remove non-unique elements
        self.interfacial_atoms_ind = list(set(self.interfacial_atoms_ind))

    def _translate_atom_1st_order(self, atom_coords, cell_vector, label, penalty_coords, count):
        """

        Parameters
        ----------
        atom_coords :
            
        cell_vector :
            
        label :
            
        penalty_coords :
            
        count :
            

        Returns
        -------

        """

        try_coords = atom_coords + cell_vector

        if not is_in_coords(try_coords, penalty_coords) and \
                not is_in_coords(try_coords, np.array(list(self.virtual_and_interfacial_atoms.values()))):
            self.virtual_and_interfacial_atoms.update({"*_" + str(count) + label: try_coords})
            count += 1

        return count

    def _translate_atom_2d_order(self, atom_coords, cell_vector, label, penalty_coords, count):
        """

        Parameters
        ----------
        atom_coords :
            
        cell_vector :
            
        label :
            
        penalty_coords :
            
        count :
            

        Returns
        -------

        """

        for vec in self.pcv:

            try_coords = atom_coords + cell_vector + vec

            if not is_in_coords(try_coords, penalty_coords) and \
                    not is_in_coords(try_coords, np.array(list(self.virtual_and_interfacial_atoms.values()))):

                self.virtual_and_interfacial_atoms.update({"**_" + str(count) + label: try_coords})
                count += 1

            try_coords = atom_coords + cell_vector - vec

            if not is_in_coords(try_coords, penalty_coords) and \
                    not is_in_coords(try_coords, np.array(list(self.virtual_and_interfacial_atoms.values()))):

                self.virtual_and_interfacial_atoms.update({"**_" + str(count) + label: try_coords})
                count += 1

        return count

    def get_neighbours(self, query):
        """

        Parameters
        ----------
        query :
            

        Returns
        -------

        """

        ans = self._get_neighbours(query)

        ans1 = []

        for item in zip(ans[0], ans[1]):
            if self._nn_distance * 0.1 < item[0] < self._nn_distance and \
                    list(self.virtual_and_interfacial_atoms.keys())[item[1]].startswith("*"):
                ans1.append(item[1])

        return ans1

    def atom_classifier(self, coords, leads):
        """

        Parameters
        ----------
        coords :
            
        leads :
            

        Returns
        -------

        """

        distance_to_surface1 = np.inner(coords - self.shift, leads) / np.linalg.norm(leads)
        distance_to_surface2 = np.inner(coords - self.shift - leads, leads) / np.linalg.norm(leads)

        flag = None

        if distance_to_surface1 < 0:
            flag = 'L'
        if distance_to_surface2 >= 0:
            flag = 'R'

        return flag


if __name__ == '__main__':

    sd = StructDesignerXYZ(xyz='/home/mk/TB_project/input_samples/SiNW2.xyz')
    sd._sort([np.argmin(np.array(list(sd.atom_list.values()))[:, 2])],
             [np.argmax(np.array(list(sd.atom_list.values()))[:, 2])])
    print("Done!")



from __future__ import absolute_import
from pkg_resources import get_distribution

__version__ = get_distribution('nano-net').version
__author__ = "Mike Klymenko"
__email__ = "mike.klymenko@rmit.edu.au"

from .orbitals import Orbitals
from .hamiltonian import Hamiltonian
from .hamiltonian_sparse import HamiltonianSp
from .aux_functions import get_k_coords, yaml_parser
from .hamiltonian_initializer import set_tb_params, initializer

from .reduced_mode_space import reduce_mode_space, bs_vs_e, bs
import logging
from pyfiglet import Figlet

logging.basicConfig(format='%(message)s', level=logging.INFO)
logging.StreamHandler(stream=None)
logging.info(Figlet(font='standard').renderText('NanoNet'))
logging.info("Vesion " + __version__)


import numpy as np


def get_high_symmetry_point_in_cartesian_space(reciprocal_lattice_vectors, coordinate_of_high_symmetry_point):
    """

    Parameters
    ----------
    reciprocal_lattice_vectors :
        
    coordinate_of_high_symmetry_point :
        

    Returns
    -------

    """
    cartesian_coordinate_in_reciprocal_space = np.squeeze(np.dot(coordinate_of_high_symmetry_point,
                                                                 reciprocal_lattice_vectors)).tolist()
    return cartesian_coordinate_in_reciprocal_space


PI = np.pi

# Lattice constants
a_si = 5.50
a_bi = 4.5332
c_bi = 11.7967
g_bi = 1.3861
gamma_bi = 0.240385652727133

SPECIAL_K_POINTS_SI = {
    'GAMMA': [0, 0, 0],
    'X': [0, 2 * PI / a_si, 0],
    'L': [PI / a_si,  PI / a_si,  PI/ a_si],
    'W': [PI / a_si,  2 * PI / a_si,  0],
    'U': [PI / (2 * a_si), 2 * PI / a_si, PI/ (2 * a_si)],
    'K': [3 * PI / (2 * a_si), 3 * PI / (2 * a_si), 0]
}


reciprocal_lattice_vectors_bi = g_bi * np.array([[-1.0, (-np.sqrt(3.0) / 3.0), (a_bi / c_bi)],
                                                  [1.0, (-np.sqrt(3.0) / 3.0), (a_bi / c_bi)],
                                                  [0.0, (2.0 * np.sqrt(3.0) / 3.0), (a_bi / c_bi)]])

SPECIAL_K_POINTS_BI = {
    'LAMBDA': get_high_symmetry_point_in_cartesian_space(reciprocal_lattice_vectors_bi, [0.25, 0.25, 0.25]),
    'GAMMA': get_high_symmetry_point_in_cartesian_space(reciprocal_lattice_vectors_bi, [0.00, 0.00, 0.00]),
    'T': get_high_symmetry_point_in_cartesian_space(reciprocal_lattice_vectors_bi, [0.50, 0.50, 0.50]),
    'L': get_high_symmetry_point_in_cartesian_space(reciprocal_lattice_vectors_bi, [0.00, 0.50, 0.00]),
    'X': get_high_symmetry_point_in_cartesian_space(reciprocal_lattice_vectors_bi, [0.50, 0.50, 0.00]),
    'K': get_high_symmetry_point_in_cartesian_space(reciprocal_lattice_vectors_bi, [((0.50 * gamma_bi) + 0.25), (0.75 -
                                                                                    (0.50 * gamma_bi)), 0.00]),
    'W': get_high_symmetry_point_in_cartesian_space(reciprocal_lattice_vectors_bi, [0.50, (1.00 - gamma_bi), gamma_bi]),
    'U': get_high_symmetry_point_in_cartesian_space(reciprocal_lattice_vectors_bi, [((0.50 * gamma_bi) + 0.25), (1.00 -
                                                                                    gamma_bi), ((0.50 * gamma_bi) +
                                                                                    0.25)])
}


"""
The module contains auxiliary global variables.
"""
import numpy as np


PI = 3.14159

# The codes for labeling the orbitals

ORBITAL_QN = {0: 's', 1: 'p', 2: 'd'}

M_QN = {0: 'sigma', 1: 'pi', 2: 'delta'}

# Coordinates of high-symmetry points in the Brillouin zone
# of the diamond-type crystal lattice

a_si = 5.50

SPECIAL_K_POINTS_SI = {
    'GAMMA': [0, 0, 0],
    'X': [0, 2 * PI / a_si, 0],
    'L': [PI / a_si,  PI / a_si,  PI / a_si],
    'W': [PI / a_si,  2 * PI / a_si,  0],
    'U': [PI / (2 * a_si), 2 * PI / a_si, PI / (2 * a_si)],
    'K': [3 * PI / (2 * a_si), 3 * PI / (2 * a_si), 0]
}

a_bi = 4.5332
c_bi = 11.7967
b_bi = a_bi / c_bi
gamma_bi = 0.2303

recip_lat_basis = np.array([[-1, -np.sqrt(3.0)/3.0, b_bi],
                             [1, -np.sqrt(3.0)/3.0, b_bi],
                             [0, 2*np.sqrt(3.0)/3.0, b_bi]]) * 2 * PI / a_bi

SPECIAL_K_POINTS_BI = {
    'GAMMA': (recip_lat_basis * np.array([0, 0, 0]).T).tolist(),
    'X': (recip_lat_basis * np.array([0.5, 0.5, 0.0]).T).tolist(),
    'T': (recip_lat_basis * np.array([0.5, 0.5, 0.5]).T).tolist(),
    'L': (recip_lat_basis * np.array([0.0, 0.5, 0.0]).T).tolist(),
    'W': (recip_lat_basis * np.array([gamma_bi, 1.0-gamma_bi, 0.5]).T).tolist(),
    'K': (recip_lat_basis * np.array([1.0/4+0.5*gamma_bi, 3.0/4-0.5*gamma_bi, 0.0]).T).tolist(),
    'U': (recip_lat_basis * np.array([0.5*gamma_bi+0.25, 1.0-gamma_bi, 0.5*gamma_bi+0.25]).T).tolist()
}

# aaa = 3.289
#
# SPECIAL_K_POINTS_BI = {
#     'GAMMA': [0, 0, 0],
#     'X': [0.5 * PI/aaa, 0.5 * PI/aaa, 0.5 * PI/aaa],
#     'T': (recip_lat_basis * np.array([0.5, 0.5, 0.5]).T).tolist(),
#     'L': (recip_lat_basis * np.array([0.0, 0.5, 0.0]).T).tolist(),
#     'W': (recip_lat_basis * np.array([gamma_bi, 1.0-gamma_bi, 0.5]).T).tolist(),
#     'K': (recip_lat_basis * np.array([1.0/4+0.5*gamma_bi, 3.0/4-0.5*gamma_bi, 0.0]).T).tolist(),
#     'U': (recip_lat_basis * np.array([0.5*gamma_bi+0.25, 1.0-gamma_bi, 0.5*gamma_bi+0.25]).T).tolist()
# }


"""
The module contains functions that computes Green's functions and their poles
"""
from __future__ import print_function, division
import pickle
import os.path
import numpy as np
from nanonet.negf.greens_functions import surface_greens_function_poles


# def object_function(vec, energy, initial_basis, extended_basis, h_0, h_0_reduced):
#
#     for eee in energy:
#
#         aaa = 1 + \
#               extended_basis.H *\
#               h_0 *\
#               initial_basis *\
#               np.pinv((z - h_0_reduced)**2) *\
#               initial_basis.H *\
#               h_0 *\
#               extended_basis
#
#         bbb = 1* z -\
#               extended_basis.H *\
#               h_0 *\
#               extended_basis - \
#               extended_basis.H * \
#               h_0 * \
#               initial_basis * \
#               np.pinv(z - h_0_reduced) * \
#               initial_basis.H * \
#               h_0 * \
#               extended_basis
#
#     return fff
def object_function1(vec, energy, init_basis, extended_basis, h_l, h_0, h_r, num_of_states):
    """

    Parameters
    ----------
    vec :
        
    energy :
        
    init_basis :
        
    extended_basis :
        
    h_l :
        
    h_0 :
        
    h_r :
        
    num_of_states :
        

    Returns
    -------

    """

    extended_basis1 = np.array(1.0 / np.sqrt(np.dot(vec, vec.conj().T))) * np.array(extended_basis * vec.T)
    extended_basis1 = np.hstack((init_basis, extended_basis1))

    h_l_reduced = np.dot(np.dot(extended_basis1.conj().T, h_l), extended_basis1)
    h_0_reduced = np.dot(np.dot(extended_basis1.conj().T, h_0), extended_basis1)
    h_r_reduced = np.dot(np.dot(extended_basis1.conj().T, h_r), extended_basis1)

    _, _, num_of_states1 = bs_vs_e(energy, h_l_reduced, h_0_reduced, h_r_reduced)

    print(num_of_states1 - num_of_states, ' : ', vec)

    return num_of_states1 - num_of_states + (np.dot(vec, vec.conj().T) - 1.0) ** 2


def object_function(vec, energy, init_basis, extended_basis, h_l, h_0, h_r, num_of_states):
    """

    Parameters
    ----------
    vec :
        
    energy :
        
    init_basis :
        
    extended_basis :
        
    h_l :
        
    h_0 :
        
    h_r :
        
    num_of_states :
        

    Returns
    -------

    """

    extended_basis1 = np.array(1.0 / np.sqrt(np.dot(vec, vec.conj().T))) * np.array(extended_basis * vec.T)
    extended_basis1 = np.hstack((init_basis, extended_basis1))

    h_l_reduced = np.dot(np.dot(extended_basis1.conj().T, h_l), extended_basis1)
    h_0_reduced = np.dot(np.dot(extended_basis1.conj().T, h_0), extended_basis1)
    h_r_reduced = np.dot(np.dot(extended_basis1.conj().T, h_r), extended_basis1)

    _, _, num_of_states1 = bs_vs_e(energy, h_l_reduced, h_0_reduced, h_r_reduced)

    print(num_of_states1 - num_of_states, ' : ', vec)

    return num_of_states1 - num_of_states + (np.dot(vec, vec.conj().T) - 1.0) ** 2


def bs(E, h_l, h_0, h_r):
    """

    Parameters
    ----------
    E :
        
    h_l :
        
    h_0 :
        
    h_r :
        

    Returns
    -------

    """

    vals, vects = surface_greens_function_poles(h_l, h_0 - E * np.identity(h_0.shape[0]), h_r)
    vals = np.diag(vals)

    vals_for_plot = vals.copy()

    acc = 0.001

    vals_for_plot = np.angle(vals_for_plot)
    vals_for_plot[np.abs(np.abs(vals) - 1.0) > acc] = np.nan
    inds = np.where(np.abs(np.abs(vals) - 1.0) <= acc)[0]
    vects = vects[:, inds]
    vals = np.angle(vals[inds])

    return vals, vects, vals_for_plot


def bs_vs_e(energy, h_l, h_0, h_r):
    """

    Parameters
    ----------
    energy :
        
    h_l :
        
    h_0 :
        
    h_r :
        

    Returns
    -------

    """

    init_basis = []
    vals_for_plot = []

    for E in energy:
        # print(E)
        _, vec, val_for_plot = bs(E, h_l, h_0, h_r)
        vals_for_plot.append(val_for_plot)
        if vec.size > 0:
            init_basis.append(vec)

    vals_for_plot = np.array(vals_for_plot)

    num_of_states = vals_for_plot[::4, :].size - np.count_nonzero(np.isnan(vals_for_plot[::4, :]))

    if len(init_basis) > 0:
        init_basis = np.hstack(tuple(init_basis))

    return init_basis, vals_for_plot, num_of_states


def reduce_mode_space(energy, h_l, h_0, h_r, thr, input_file=""):
    """

    Parameters
    ----------
    energy :
        
    h_l :
        
    h_0 :
        
    h_r :
        
    thr :
        
    input_file :
         (Default value = "")

    Returns
    -------

    """

    # energy = np.linspace(2.0, 3.7, 50)

    if os.path.isfile(input_file):
        input_file = os.path.dirname(input_file)

    label = '_' + "{0:.2f}".format(np.min(energy)) + '_' + "{0:.2f}".format(np.max(energy)) + '_' + str(len(energy))
    first_file = os.path.join(input_file, 'init_basis'+label+'.pkl')
    second_file = os.path.join(input_file, 'vals_for_plot'+label+'.pkl')

    if os.path.isfile(first_file) and os.path.isfile(second_file):
        # unpickle
        with open(first_file, 'rb') as infile:
            init_basis = pickle.load(infile)
        with open(second_file, 'rb') as infile:
            vals_for_plot = pickle.load(infile)

        num_of_states = vals_for_plot[::4, :].size - np.count_nonzero(np.isnan(vals_for_plot[::4, :]))
    else:
        init_basis, vals_for_plot, num_of_states = bs_vs_e(energy, h_l, h_0, h_r)
        # pickle
        with open(first_file, 'wb') as outfile:
            pickle.dump(init_basis, outfile)
        with open(second_file, 'wb') as outfile:
            pickle.dump(vals_for_plot, outfile)

    # orthogonalize initial basis

    eee, vvv = np.linalg.eig(np.dot(init_basis.conj().T, init_basis))
    init_basis = init_basis.dot(vvv).dot(np.diag(1.0/np.sqrt(eee)))
    init_basis = init_basis[:, np.where(eee > thr)[0]]

    # test reduced mode space

    h_l_reduced = np.dot(np.dot(init_basis.conj().T, h_l), init_basis)
    h_0_reduced = np.dot(np.dot(init_basis.conj().T, h_0), init_basis)
    h_r_reduced = np.dot(np.dot(init_basis.conj().T, h_r), init_basis)

    _, vals_for_plot_1, num_of_states1 = bs_vs_e(energy, h_l_reduced, h_0_reduced, h_r_reduced)

    # while num_of_states != num_of_states1:
    #
    #     extended_basis = (1.0 - init_basis * init_basis.H) * h_0 * init_basis
    #     extended_basis1 = (1.0 - init_basis * init_basis.H) * (h_l + h_r) * init_basis
    #     extended_basis = np.matrix(np.hstack((extended_basis, extended_basis1)))
    #     eee, vvv = np.linalg.eig(extended_basis.H * extended_basis)
    #     extended_basis = extended_basis * vvv * np.matrix(np.diag(1.0 / np.sqrt(eee)))
    #     extended_basis = extended_basis[:, np.where(eee > thr)[0]]
    #     x0 = 0.5*np.ones(extended_basis.shape[1])
    #     res = minimize(object_function,
    #                    x0,
    #                    args=(energy, init_basis, extended_basis, h_l, h_0, h_r, num_of_states),
    #                    method='COBYLA', options={'maxiter': 300})
    #     vec = np.matrix(res.x)
    #
    #     extended_basis = np.matrix(np.array(1.0 / np.sqrt(vec * vec.H)) * np.array(extended_basis * vec.T))
    #     init_basis = np.hstack((init_basis, extended_basis))
    #
    #     # test reduced mode space
    #     h_l_reduced = init_basis.H * h_l * init_basis
    #     h_0_reduced = init_basis.H * h_0 * init_basis
    #     h_r_reduced = init_basis.H * h_r * init_basis
    #
    #     _, vals_for_plot2, num_of_states1 = bs_vs_e(energy, h_l_reduced, h_0_reduced, h_r_reduced)

    return h_l_reduced, h_0_reduced, h_r_reduced, vals_for_plot, init_basis



import numpy as np
from scipy.linalg import eig


class CFR(object):
    """ """

    val_inf = 1.0e4

    def __init__(self, cutoff=0):
        self.cutoff = cutoff
        if cutoff == 0:
            self.fd_poles_coords = []
            self.fd_poles = []
            self.num_poles = 0
        else:
            self.gen_poles_and_residues(self.cutoff)

    def fd_approximant(self, energy):
        """

        Parameters
        ----------
        energy :
            

        Returns
        -------

        """
        arg = np.array(energy)
        ans = np.zeros(arg.shape) + 0.5

        for j in range(self.num_poles):
            if self.fd_poles_coords[j] > 0:
                ans = ans - self.fd_poles[j] / (arg - 1j / self.fd_poles_coords[j]) - \
                      self.fd_poles[j] / (arg + 1j / self.fd_poles_coords[j])

        return ans

    def fd_approximant_diff(self, energy):
        """

        Parameters
        ----------
        energy :
            

        Returns
        -------

        """

        arg = np.array(energy)
        ans = np.zeros(arg.shape) + 0.5 * 0

        for j in range(self.num_poles):
            if self.fd_poles_coords[j] > 0:
                ans = ans - self.fd_poles[j] / (arg - 1j / self.fd_poles_coords[j]) ** 2 - \
                      self.fd_poles[j] / (arg + 1j / self.fd_poles_coords[j]) ** 2

        return ans

    def gen_poles_and_residues(self, cutoff=50):
        """Compute positions of poles and their residuals for the Fermi-Dirac function

        Parameters
        ----------
        cutoff :
            cutoff energy (Default value = 50)

        Returns
        -------

        """

        self.cutoff = cutoff

        b_mat = [1 / (2.0 * np.sqrt((2 * (j + 1) - 1) * (2 * (j + 1) + 1))) for j in range(0, cutoff - 1)]
        b_mat = np.diag(b_mat, -1) + np.diag(b_mat, 1)

        poles, residues = eig(b_mat)

        residues = 0.25 * np.array([np.abs(residues[0, j]) ** 2 / (poles[j] ** 2) for j in range(residues.shape[0])])

        self.fd_poles_coords = poles
        self.fd_poles = residues
        self.num_poles = len(self.fd_poles_coords)

    def get_poles_and_residues(self):
        """ """
        return self.fd_poles_coords, self.fd_poles

    def integrate(self, gf, ef=0, tempr=300, zero_moment=-1):
        """

        Parameters
        ----------
        gf :
            param ef:
        tempr :
            param zero_moment: (Default value = 300)
        ef :
             (Default value = 0)
        zero_moment :
             (Default value = -1)

        Returns
        -------

        """

        if zero_moment == -1:
            zero_moment = 1j * CFR.val_inf * gf(1j * CFR.val_inf)

        ans = 0
        betha = 1.0 / (8.617333262145e-5 * tempr)

        for j in range(self.num_poles):
            if np.real(self.fd_poles_coords[j]) > 0:
                aaa = ef + 1j / self.fd_poles_coords[j] / betha
                ans = ans + 4 * 1j / betha * gf(aaa) * self.fd_poles[j]

        return np.real(zero_moment + np.imag(ans))

    def genetate_integration_points(self, ef, tempr):
        """

        Parameters
        ----------
        ef :
            
        tempr :
            

        Returns
        -------

        """
        ans = []
        betha = 1.0 / (8.617333262145e-5 * tempr)

        for j in range(self.num_poles):
            ans.append(ef + 1j / self.fd_poles_coords[j] / betha)

        return np.array(ans)

    def integrate1(self, gf_vals, tempr, zero_moment=0):
        """

        Parameters
        ----------
        gf_vals :
            param tempr:
        zero_moment :
            return: (Default value = 0)
        tempr :
            

        Returns
        -------

        """

        assert len(gf_vals) == len(self.fd_poles_coords)

        ans = 0
        betha = 1.0 / (8.617333262145e-5 * tempr)

        for j in range(self.num_poles):
            if np.real(self.fd_poles_coords[j]) > 0:
                ans = ans + 4 * 1j / betha * gf_vals[j] * self.fd_poles[j]

        return np.real(zero_moment + np.imag(ans))



import unittest
import doctest
import nanonet.tb.block_tridiagonalization as btd
suite = doctest.DocTestSuite(btd)
unittest.TextTestRunner(verbosity=2).run(suite)

if __name__ == "__main__":
    unittest.TextTestRunner(verbosity=2).run(suite)



"""
This function generates the weights required for the discrete
Hilbert transform. Re[G] = -Im[G]*phi, Im[G] = Re[G]*phi.

The reason for a discrete hilbert transform is that the common
fast fourier transform method erroneously pins the the real
part at the edges leading to significant deviation from the
true result. For a given energy grid, this method generates
a matrix to which you multiply your signal, after which the
method returns the Hilbert transform of that function. This
method assumes that the functions are localised to the domain
window, if there is an approximately constant tail in the
positive/negative direction, an additional term must be added.
If the function is actually periodic we recommend you use the
FFT method, as it will be better suited to your problem.

If used, please cite
Jesse Vaitkus' thesis 2020
"""


import numpy as np


def make_hilbert_weights(E, **kwargs):
    """This function generates the weights required for the discrete
       Hilbert transform. This method assumes that each site can be
       describe using linearly interpolating basis functions. Other
       basis functions are planned.

       Re[G] = -Im[G]*phi, Im[G] = Re[G]*phi.

        Parameters
        ----------
        E : numpy array
            Energy grid to compute weights for
        
        kwargs : dict
              additional named arguments:
                eta : float (Default value = sqrt(eps) )
                      Imaginary part to be used for stability, should
                      be as small as possible without breaking

        Returns
        -------
        numpy.matrix

            #Guide for usage.

            A user provides an energy grid and the method returns an
            array phi such that Re[G] = -Im[G]*phi, Im[G] = Re[G]*phi.
        """

    eta = kwargs.get('eta', np.sqrt(np.finfo(float).eps))

    numE = np.size(E)
    dE = (E[-1] - E[0])/(numE - 1)  # This is the average separation of points,
    Eneg = E[0] - dE                # which we use to generate the ghost points
    Epos = E[-1] + dE               # Eneg/Epos to handle the domain boundaries.

    phi = np.empty( (numE, numE), dtype=complex)

    phi[0, :] = (E + 1j*eta - Eneg)/(E[0] - Eneg)*np.log((Eneg - E - 1j*eta)/(E[0] - E - 1j*eta)) + \
                (E + 1j*eta - E[1])/(E[0] - E[1])*np.log((E + 1j*eta - E[0])/(E + 1j*eta - E[1]))

    for ii in range(1, numE-1):
        phi[ii, :] = (E + 1j * eta - E[ii-1])/(E[ii] - E[ii-1])*np.log((E[ii-1] - E - 1j*eta)/(E[ii] - E - 1j*eta)) + \
                     (E + 1j * eta - E[ii+1])/(E[ii] - E[ii+1])*np.log((E + 1j*eta - E[ii])/(E + 1j * eta - E[ii+1]))

    phi[-1, :] = (E + 1j * eta - E[-2]) / (E[-1] - E[-2]) * np.log((E[-2] - E - 1j * eta) / (E[-1] - E - 1j * eta)) + \
                 (E + 1j * eta - Epos) / (E[-1] - Epos) * np.log((E + 1j * eta - E[-1]) / (E + 1j * eta - Epos))

    return np.real(phi)/np.pi  # Discard imaginary part used for stability in computation.



"""
The module contains abstract interfaces of the classes.
The interfaces are aimed to be schemas for further classes implementations.
Following these schemas will ensure compatibility of the code with the entire project.
"""
from abc import ABCMeta, abstractmethod
from future.utils import with_metaclass
import numpy as np


class AbstractStructureDesigner(with_metaclass(ABCMeta, object)):
    """The class is an abstraction for the list of atomic coordinates."""

    def __init__(self):

        self._nn_distance = None
        self._kd_tree = None

    def _get_neighbours(self, query):
        """

        Parameters
        ----------
        query :
            

        Returns
        -------

        """

        if isinstance(query, list) or isinstance(query, np.ndarray):
            ans = self._kd_tree.query(query,
                                      k=25,
                                      distance_upper_bound=self._nn_distance)
        elif isinstance(query, int):

            query = list(self.atom_list.items())[query][1]

            if self._kd_tree.data.shape[1] > 3:
                query = np.append(query, 0)

            ans = self._kd_tree.query(query,
                                      k=25,
                                      distance_upper_bound=self._nn_distance)
        elif isinstance(query, str):
            ans = self._kd_tree.query(self.atom_list[query],
                                      k=25,
                                      distance_upper_bound=self._nn_distance)
        else:
            raise TypeError('Wrong input type for query')

        return ans

    @abstractmethod
    def get_neighbours(self, query):
        """

        Parameters
        ----------
        query :
            

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def atom_list(self):
        """ """
        pass


class AbstractBasis(with_metaclass(ABCMeta, object)):
    """The class contains information about sets of quantum numbers and
    dimensionality of the Hilbert space.
    It is also equipped with the member functions translating quantum numbers
    into a raw index and vise versa.

    Parameters
    ----------

    Returns
    -------

    """

    @abstractmethod
    def qn2ind(self, qn):
        """The member function trasform a dictionary of quantum numbers into a matrix index

        Parameters
        ----------
        qn :
            type qn:
            
            :return ind:   index

        Returns
        -------

        """
        pass

    @abstractmethod
    def ind2qn(self, ind):
        """

        Parameters
        ----------
        ind :
            

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def orbitals_dict(self):
        """ """
        pass



"""
The module contains functions facilitating setting tight-binding parameters and
initializing Hamiltonian objects from a Python dictionary.
"""

from __future__ import absolute_import
import sys
import numpy as np
from nanonet.tb.orbitals import Orbitals
from nanonet.tb import tb_params as dme
from nanonet.tb.hamiltonian import Hamiltonian
from nanonet.tb.hamiltonian_sparse import HamiltonianSp


def set_tb_params(**kwargs):
    """Initialize a set of the user-defined tight-binding parameters.

    Parameters
    ----------
    **kwargs :
        

    Returns
    -------

    
    """
    for item in kwargs:
        if item.startswith('PARAMS_') or item.startswith('OV_'):
            setattr(dme, item, kwargs[item])


def initializer(**kwargs):
    """Creates a Hamiltonian object from a set of parameters stored in a Python dictionary.
    
    This functions is used by CLI scripts to create Hamiltonian objects
    from a configuration file (normally in a yaml format) which is previously
    parsed into a Python dictionary data structure.

    Parameters
    ----------
    kwargs : dict
        Dictionary of parameters needed to make a Hamiltonian object.
    **kwargs :
        

    Returns
    -------

    
    """
    set_tb_params(**kwargs)
    Orbitals.orbital_sets = kwargs.get('orbital_sets', {'Si': 'SiliconSP3D5S', 'H': 'HydrogenS'})
    sys.modules[__name__].VERBOSITY = kwargs.get('VERBOSITY', 1)

    xyz = kwargs.get('xyz', {})
    nn_distance = kwargs.get('nn_distance', 2.7)
    sparse = kwargs.get('sparse', 0)
    sigma = kwargs.get('sigma', 1.1)
    num_eigs = kwargs.get('num_eigs', 14)

    if sparse:
        h = HamiltonianSp(xyz=xyz, nn_distance=nn_distance, sigma=sigma, num_eigs=num_eigs)
    else:
        h = Hamiltonian(xyz=xyz, nn_distance=nn_distance)

    h.initialize()

    primitive_cell = kwargs.get('primitive_cell', [0, 0, 0])

    if np.sum(np.abs(np.array(primitive_cell))) > 0:
        h.set_periodic_bc(primitive_cell=primitive_cell)

    return h



import numpy as np
import nanonet.tb as tb
from test.test_hamiltonian_module import expected_bulk_silicon_band_structure


def test_simple_atomic_chain():
    """ """
    site_energy = -1.0
    coupling = -1.0
    l_const = 1.0

    a = tb.Orbitals('A')
    a.add_orbital(title='s', energy=-1, )

    xyz_file = """1
    H cell
    A       0.0000000000    0.0000000000    0.0000000000
    """
    tb.set_tb_params(PARAMS_A_A={'ss_sigma': -1.0})
    h = tb.HamiltonianSp(xyz=xyz_file, nn_distance=1.1)
    h.initialize()

    PRIMITIVE_CELL = [[0, 0, l_const]]
    h.set_periodic_bc(PRIMITIVE_CELL)

    num_points = 10
    kk = np.linspace(0, 3.14 / l_const, num_points, endpoint=True)

    band_structure = []

    for jj in range(num_points):
        vals, _ = h.diagonalize_periodic_bc([0.0, 0.0, kk[jj]])
        band_structure.append(vals)

    band_structure = np.array(band_structure)

    desired_value = site_energy + 2 * coupling * np.cos(l_const * kk)
    np.testing.assert_allclose(band_structure, desired_value[:, np.newaxis], atol=1e-9)


def test_atomic_chain_two_kinds_of_atoms():
    """ """
    site_energy1 = -1.0
    site_energy2 = -2.0
    coupling = -1.0
    l_const = 2.0

    a = tb.Orbitals('A')
    a.add_orbital(title='s', energy=site_energy1, )
    b = tb.Orbitals('B')
    b.add_orbital(title='s', energy=site_energy2, )

    xyz_file = """2
    H cell
    A       0.0000000000    0.0000000000    0.0000000000
    B       0.0000000000    0.0000000000    1.0000000000
    """
    tb.set_tb_params(PARAMS_A_B={'ss_sigma': coupling})
    h = tb.HamiltonianSp(xyz=xyz_file, nn_distance=1.1)
    h.initialize()

    PRIMITIVE_CELL = [[0, 0, l_const]]
    h.set_periodic_bc(PRIMITIVE_CELL)

    num_points = 10
    kk = np.linspace(0, 3.14 / 2, num_points, endpoint=True)

    band_structure = []

    for jj in range(num_points):
        vals, _ = h.diagonalize_periodic_bc([0.0, 0.0, kk[jj]])
        band_structure.append(vals)

    band_structure = np.array(band_structure)
    desired_value = np.zeros(band_structure.shape)

    b = site_energy1 + site_energy2
    c = site_energy1 * site_energy2 - (2.0 * coupling * np.cos(0.5 * kk * l_const)) ** 2
    desired_value[:, 0] = 0.5 * (b - np.sqrt(b ** 2 - 4.0 * c))
    desired_value[:, 1] = 0.5 * (b + np.sqrt(b ** 2 - 4.0 * c))

    np.testing.assert_allclose(band_structure, desired_value, atol=1e-9)


def test_bulk_silicon():
    """ """
    a_si = 5.50
    PRIMITIVE_CELL = [[0, 0.5 * a_si, 0.5 * a_si],
                      [0.5 * a_si, 0, 0.5 * a_si],
                      [0.5 * a_si, 0.5 * a_si, 0]]

    tb.Orbitals.orbital_sets = {'Si': 'SiliconSP3D5S'}

    xyz_file = """2
    Si2 cell
    Si1       0.0000000000    0.0000000000    0.0000000000
    Si2       1.3750000000    1.3750000000    1.3750000000
    """

    h = tb.HamiltonianSp(xyz=xyz_file, nn_distance=2.5)
    h.initialize()
    h.set_periodic_bc(PRIMITIVE_CELL)

    sym_points = ['L', 'GAMMA', 'X']
    num_points = [10, 25]
    k = tb.get_k_coords(sym_points, num_points, 'Si')
    band_sructure = []

    vals = np.zeros((sum(num_points), h.num_eigs), dtype=complex)

    for jj, item in enumerate(k):
        vals[jj, :], _ = h.diagonalize_periodic_bc(item)

    band_structure = np.real(np.array(vals))
    np.testing.assert_allclose(band_structure, expected_bulk_silicon_band_structure()[:,:h.num_eigs], atol=1e-4)


if __name__ == '__main__':
    # test_simple_atomic_chain()
    # test_atomic_chain_two_kinds_of_atoms()
    test_bulk_silicon()



#!/usr/bin/env python
from __future__ import print_function
from __future__ import absolute_import
try:
    from mpi4py import MPI

    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()
except ImportError:
    rank = 0
    size = 1
import nanonet.tb as tb
from nanonet.tb.tb_script import create_parser, preprocess_data, postprocess_data


def main1(param_file, k_points_file, xyz, show, save, code_name):
    """

    Parameters
    ----------
    param_file :
        
    k_points_file :
        
    xyz :
        
    show :
        
    save :
        
    code_name :
        

    Returns
    -------

    """

    params, wave_vector, code_name = preprocess_data(param_file, k_points_file, xyz, code_name)

    # initialize Hamiltonian
    hamiltonian = tb.initializer(**params)

    # compute band structure
    band_structure = []

    for j, jj in enumerate(wave_vector):
        if j % size != rank:
            continue

        vals, vects = hamiltonian.diagonalize_periodic_bc(jj)
        band_structure.append({'id': j, 'wave_vector': jj, 'eigenvalues': vals, 'eigenvectors': vects})
        print('#{} '.format(j), " ".join(['{:.3f} '.format(element) for element in vals]))

    if size > 1:
        band_structure = comm.reduce(band_structure, root=0)

    if rank == 0:
        postprocess_data(wave_vector, band_structure, show, save, code_name)

    return 0


def main():
    """ """

    parser = create_parser()
    args = parser.parse_args()
    main1(args.param_file, args.k_points_file, args.xyz, args.show, args.save, args.code_name)

    return 0


if __name__ == '__main__':

    main()



"""
The module contains a library of classes facilitating computations of Hamiltonian matrices.
"""
from __future__ import print_function, division
from __future__ import absolute_import
from collections import OrderedDict
from functools import reduce
import logging
import inspect
from operator import mul
import numpy as np
import scipy
from nanonet.tb.abstract_interfaces import AbstractBasis
from nanonet.tb.structure_designer import StructDesignerXYZ, CyclicTopology
from nanonet.tb.diatomic_matrix_element import me
from nanonet.tb.orbitals import Orbitals
from nanonet.tb.aux_functions import dict2xyz
from nanonet.tb.block_tridiagonalization import find_nonzero_lines, split_into_subblocks_optimized, cut_in_blocks, split_into_subblocks
import nanonet.verbosity as verbosity


unique_distances = set()


class BasisTB(AbstractBasis, StructDesignerXYZ):
    """The class contains information about sets of quantum numbers and
    dimensionality of the Hilbert space.
    It is also equipped with the member functions translating quantum numbers
    into a matrix index and vise versa using a set of index offsets.

    Examples
    --------
    >>> from nanonet.verbosity import set_verbosity
    >>> import nanonet.tb as tb
    >>> set_verbosity(0)
    >>> orb = tb.Orbitals('A')
    >>> orb.add_orbital(title='s', energy=-1)
    >>> orb.add_orbital(title='1s', energy=0)
    >>> tb.Orbitals('B').add_orbital(title='s', energy=0)
    >>> xyz = '''2
    ... Two atoms
    ... A1 0 0 0
    ... B1 0 0 1'''
    >>> basis = tb.hamiltonian.BasisTB(xyz=xyz)
    >>> print(basis.basis_size)
    3
    >>> print(basis.atom_list['B1'])
    [0. 0. 1.]
    >>> print(basis.qn2ind({'atoms': 0, 'l': 0}))
    0
    >>> print(basis.qn2ind({'atoms': 0, 'l': 1}))
    1
    >>> print(basis.qn2ind({'atoms': 1, 'l': 0}))
    2
    >>> print(type(basis.orbitals_dict['A']))
    <class 'nanonet.tb.orbitals.Orbitals'>
    """

    def __init__(self, **kwargs):

        # parent class StructDesignerXYZ stores atom list initialized from xyz-file
        super(BasisTB, self).__init__(**kwargs)

        # each entry of the dictionary stores a label of the atom species as a key and
        # corresponding Atom object as a value. Each atom object contains infomation about number,
        # energy and symmetry of the orbitals
        self._orbitals_dict = Orbitals.atoms_factory(list(self.num_of_species.keys()))

        # `quantum_number_lims` counts number of species and corresponding number
        # of orbitals for each; each atom kind is enumerated
        self.quantum_numbers_lims = []
        for item in list(self.num_of_species.keys()):
            self.quantum_numbers_lims.append(OrderedDict([('atoms', self.num_of_species[item]),
                                                          ('l', self.orbitals_dict[item].num_of_orbitals)]))

        # count total number of basis functions
        self.basis_size = 0
        for item in self.quantum_numbers_lims:
            self.basis_size += reduce(mul, list(item.values()))

        # compute offset index for each atom
        self._offsets = [0]
        for j in range(len(self.atom_list) - 1):
            self._offsets.append(self.orbitals_dict[list(self.atom_list.keys())[j]].num_of_orbitals)
        self._offsets = np.cumsum(self._offsets)

        # make a log
        logging.info("Basis set \n Num of species {} \n".format(self.num_of_species))
        for key, label in self._orbitals_dict.items():
            logging.info("\n {} {} ".format(key, label.generate_info()))
        logging.info("---------------------------------\n")

    def qn2ind(self, qn):
        """Computes a matrix index of an matrix element from the index of atom and the index of atomic orbital.

        Parameters
        ----------
        qn : dict
            A dictionary with two keys `atoms` and `l`, where the fist one is the atom index and
            the later is the orbital index.

        Returns
        -------
        type int
            Index of the TB matrix
        """

        qn = OrderedDict(qn)

        if list(qn.keys()) == list(self.quantum_numbers_lims[0].keys()):  # check if the input is
            # a proper set of quantum numbers
            return self._offsets[qn['atoms']] + qn['l']
        else:
            raise IndexError("Wrong set of quantum numbers")

    def ind2qn(self, ind):
        """

        Parameters
        ----------
        ind :
            

        Returns
        -------

        """
        pass  # TODO

    @property
    def orbitals_dict(self):
        """Returns the dictionary data structure of orbitals. In the dictionary"""

        class MyDict(dict):
            """ """
            def __getitem__(self, key):
                key = ''.join([i for i in key if not i.isdigit()])
                return super(MyDict, self).__getitem__(key)

        return MyDict(self._orbitals_dict)


class Hamiltonian(BasisTB):
    """Class defines a Hamiltonian matrix as well as a set of member-functions
    allowing to build, diagonalize and visualize the matrix.

    Examples
    --------
    >>> from nanonet.verbosity import set_verbosity
    >>> import nanonet.tb as tb
    >>> set_verbosity(0)
    >>> tb.Orbitals('A').add_orbital(title='s', energy=-1)
    >>> tb.Orbitals('B').add_orbital(title='s', energy=-2)
    >>> xyz_file = '''2
    ... Two atoms
    ... A1 0 0 0
    ... B1 0 0 1.5'''
    >>> tb.set_tb_params(PARAMS_A_B={'ss_sigma': 0.1})
    >>> h = tb.Hamiltonian(xyz=xyz_file, nn_distance=2.0).initialize()
    >>> h.h_matrix
    array([[-1. +0.j,  0.1+0.j],
           [ 0.1+0.j, -2. +0.j]])
    """

    def __init__(self, **kwargs):
        """

        Parameters
        ----------
        nn_distance : float, list
            Nearest neighbour distance of distances
            if the tight-binding method is beyond the first-nearest neighbour approximation (Default value = 2.39)
        radial_dep : func
             Radial dependence function (Default value = None)
        """

        nn_distance = kwargs.get('nn_distance', 2.39)
        self.int_radial_dependence = None
        nn_distance = self._set_nn_distances(nn_distance)
        self.compute_overlap = kwargs.get('comp_overlap', False)
        self.compute_angular = kwargs.get('comp_angular_dep', True)

        logging.info('The verbosity level is {}'.format(verbosity.VERBOSITY))
        logging.info('The radius of the neighbourhood is {} Ang'.format(nn_distance))
        logging.info("\n---------------------------------\n")

        kwargs['nn_distance'] = nn_distance

        if not isinstance(kwargs['xyz'], str):
            kwargs['xyz'] = dict2xyz(kwargs['xyz'])

        super(Hamiltonian, self).__init__(**kwargs)

        self._coords = None  # coordinates of sites
        self.h_matrix = None  # Hamiltonian for an isolated system
        self.ov_matrix = None  # overlap matrix for an isolated system
        self.h_matrix_bc_factor = None  # exponential Bloch factors for pbc
        self.h_matrix_bc_add = None  # additive Bloch exponentials for pbc
        self.ov_matrix_bc_add = None  # additive Bloch exponentials for pbc
        # (interaction with virtual neighbours
        # in adacent primitive cells due to pbc)

        self.h_matrix_left_lead = None
        self.h_matrix_right_lead = None
        self.k_vector = 0  # default value of the wave vector
        self.ct = None
        self.radial_dependence = None
        self.so_coupling = kwargs.get('so_coupling', 0.0)

        radial_dep = kwargs.get('radial_dep', None)

        if radial_dep is None:
            logging.info('Radial dependence function: None')
            logging.info("\n---------------------------------\n")
        else:
            logging.info('Radial dependence function:\n\n{}'.format(inspect.getsource(radial_dep)))
            logging.info("\n---------------------------------\n")

        self.radial_dependence = radial_dep

    def initialize(self):
        """Compute matrix elements of the Hamiltonian.

        Returns
        -------
        type Hamiltonian
            Returns the instance of the class Hamiltonian
        """

        self._coords = [0 for _ in range(self.basis_size)]
        # initialize Hamiltonian matrices
        self.h_matrix = np.zeros((self.basis_size, self.basis_size), dtype=complex)
        self.h_matrix_bc_add = np.zeros((self.basis_size, self.basis_size), dtype=complex)
        self.h_matrix_bc_factor = np.ones((self.basis_size, self.basis_size), dtype=complex)

        if self.compute_overlap:
            self.ov_matrix = np.zeros((self.basis_size, self.basis_size), dtype=complex)
            self.ov_matrix_bc_add = np.zeros((self.basis_size, self.basis_size), dtype=complex)

        # loop over all nodes
        for j1 in range(self.num_of_nodes):

            # find neighbours for each node
            list_of_neighbours = self.get_neighbours(j1)

            for j2 in list_of_neighbours:
                # on site interactions
                if j1 == j2:
                    for l1 in range(self.orbitals_dict[list(self.atom_list.keys())[j1]].num_of_orbitals):
                        ind1 = self.qn2ind([('atoms', j1), ('l', l1)], )
                        self.h_matrix[ind1, ind1] = self._get_me(j1, j2, l1, l1)
                        if self.compute_overlap:
                            self.ov_matrix[ind1, ind1] = self._get_me(j1, j2, l1, l1, overlap=True)
                        self._coords[ind1] = list(self.atom_list.values())[j1]

                        if self.so_coupling != 0:
                            for l2 in range(self.orbitals_dict[list(self.atom_list.keys())[j1]].num_of_orbitals):
                                ind2 = self.qn2ind([('atoms', j1), ('l', l2)], )
                                self.h_matrix[ind1, ind2] = self._get_me(j1, j2, l1, l2)

                # nearest neighbours interaction
                else:
                    for l1 in range(self.orbitals_dict[list(self.atom_list.keys())[j1]].num_of_orbitals):
                        for l2 in range(self.orbitals_dict[list(self.atom_list.keys())[j2]].num_of_orbitals):
                            ind1 = self.qn2ind([('atoms', j1), ('l', l1)], )
                            ind2 = self.qn2ind([('atoms', j2), ('l', l2)], )

                            self.h_matrix[ind1, ind2] = self._get_me(j1, j2, l1, l2)
                            if self.compute_overlap:
                                self.ov_matrix[ind1, ind2] = self._get_me(j1, j2, l1, l2, overlap=True)

        logging.info("Unique distances: \n    {}".format("\n    ".join(unique_distances)))
        logging.info("---------------------------------\n")

        return self

    def set_periodic_bc(self, primitive_cell):
        """Set periodic boundary conditions.
        The function sets the periodic boundary conditions by creating an object of the class CyclicTopology.

        Parameters
        ----------
        primitive_cell : list
            list of vectors defining a primitive cell
        """
        if list(primitive_cell):
            self.ct = CyclicTopology(primitive_cell,
                                     list(self.atom_list.keys()),
                                     list(self.atom_list.values()),
                                     self._nn_distance)
        else:
            self.ct = None

    def diagonalize(self):
        """Diagonalize the Hamiltonian matrix for the finite isolated system
        (without periodic boundary conditions)

        Returns
        -------
        vals : numpy.ndarray
            Eigenvalues
        vects : numpy.ndarray
            Eigenvectors
        """

        vals, vects = scipy.linalg.eigh(self.h_matrix, self.ov_matrix)
        vals = np.real(vals)
        ind = np.argsort(vals)

        return vals[ind], vects[:, ind]

    def diagonalize_periodic_bc(self, k_vector):
        """Diagonalize the Hamiltonian matrix with the periodic boundary conditions
        for a certain value of the wave vector k_vector

        Parameters
        ----------
        k_vector : numpy.ndarray
            wave vector

        Returns
        -------
        vals : numpy.ndarray
            Eigenvalues
        vects : numpy.ndarray
            Eigenvectors
        """

        k_vector = list(k_vector)

        # reset previous wave vector if any
        if k_vector != self.k_vector:
            self._reset_periodic_bc()
            self.k_vector = k_vector
            self._compute_h_matrix_bc_factor()
            self._compute_h_matrix_bc_add(overlap=self.compute_overlap)

        if self.compute_overlap:
            vals, vects = scipy.linalg.eigh(self.h_matrix_bc_factor * self.h_matrix + self.h_matrix_bc_add,
                                            self.h_matrix_bc_factor * self.ov_matrix + self.ov_matrix_bc_add)
        else:
            vals, vects = np.linalg.eigh(self.h_matrix_bc_factor * self.h_matrix + self.h_matrix_bc_add)

        vals = np.real(vals)
        ind = np.argsort(vals)

        return vals[ind], vects[:, ind]

    def _set_nn_distances(self, nn_dist):
        if nn_dist is not None:
            if isinstance(nn_dist, list):
                logging.info('{} nearest neighbour interactions are taken into account.'.format(len(nn_dist)))
                logging.info("\n---------------------------------\n")
                nn_dist.sort()
                self._nn_distance = nn_dist[-1]

                def int_radial_dep(coords):
                    """
                        Step-wise radial dependence function
                    """
                    norm_of_coords = np.linalg.norm(coords)

                    ans = sum([norm_of_coords > item for item in nn_dist]) + 1

                    if norm_of_coords > nn_dist[-1]:
                        return 100
                    else:
                        return ans

                self.int_radial_dependence = int_radial_dep
            else:
                self._nn_distance = nn_dist
        else:
            logging.info('The first nearest-neighbour approximation is used.')
            logging.info("\n---------------------------------\n")

        return self._nn_distance

    def _ind2atom(self, ind):
        """

        Parameters
        ----------
        ind :
            

        Returns
        -------

        """

        return self.orbitals_dict[list(self.atom_list.keys())[ind]]

    def _get_me(self, atom1, atom2, l1, l2, coords=None, overlap=False):
        """Compute the matrix element <atom1, l1|H|l2, atom2>.
        The function is called in the member function initialize() and invokes the function
        me() from the module diatomic_matrix_element.

        Parameters
        ----------
        atom1 : int
            Atom index
        atom2 : int
            Atom index
        l1 : int
            Index of a localized basis function
        l2 : int
            Index of a localized basis function
        coords : numpy.ndarray
            Coordinates of radius vector pointing from one atom to another
            it may differ from the actual coordinates of atoms (Default value = None)
        overlap : bool
            A flag indicating that the overlap matrix element has to be computed

        Returns
        -------
        type float
            Inter-cites matrix element
        """

        # on site (pick right table of parameters for a certain atom)
        if atom1 == atom2 and coords is None:
            atom_obj = self._ind2atom(atom1)
            if l1 == l2:
                if overlap:
                    return 1.0
                else:
                    return atom_obj.orbitals[l1]['energy']
            else:
                return self._comp_so(atom_obj, l1, l2)

        # nearest neighbours (define bound type and atomic quantum numbers)
        if atom1 != atom2 or coords is not None:

            atom_kind1 = self._ind2atom(atom1)
            atom_kind2 = self._ind2atom(atom2)

            # compute radius vector pointing from one atom to another
            if coords is None:
                coords1 = np.array(list(self.atom_list.values())[atom1], dtype=float) - \
                          np.array(list(self.atom_list.values())[atom2], dtype=float)
            else:
                coords1 = coords.copy()

            norm = np.linalg.norm(coords1)

            if verbosity.VERBOSITY > 0:

                coordinates = np.array2string(norm, precision=4) + " Ang between atoms " + \
                              self._ind2atom(atom1).title + " and " + self._ind2atom(atom2).title

                if coordinates not in unique_distances:
                    unique_distances.add(coordinates)

            if self.int_radial_dependence is None:
                which_neighbour = ""
            else:
                which_neighbour = self.int_radial_dependence(norm)

            if self.radial_dependence is None:
                factor = 1.0
            else:
                factor = self.radial_dependence(norm)

            # compute directional cosines
            if self.compute_angular:
                coords1 /= norm
            else:
                coords1 = np.array([1.0, 0.0, 0.0])

            return me(atom_kind1, l1, atom_kind2, l2, coords1, which_neighbour,
                      overlap=overlap) * factor

    def _comp_so(self, atom, ind1, ind2):
        """

        Parameters
        ----------
        atom : Atom
            
        ind1 :
            
        ind2 :
            

        Returns
        -------
        type float
            Spin-orbit coupling energy
        """

        type1 = atom.orbitals[ind1]['title']
        type2 = atom.orbitals[ind2]['title']

        # quantum numbers
        l1 = atom.orbitals[ind1]['l']
        s1 = atom.orbitals[ind1]['s']
        l2 = atom.orbitals[ind2]['l']
        s2 = atom.orbitals[ind2]['s']

        if l1 == 1 and l2 == 1:

            if type1 == 'px' and type2 == 'py' and s1 == 0 and s2 == 0:
                return -1j * self.so_coupling / 3

            elif type1 == 'px' and type2 == 'pz' and s1 == 0 and s2 == 1:
                return self.so_coupling / 3

            elif type1 == 'py' and type2 == 'pz' and s1 == 0 and s2 == 1:
                return -1j * self.so_coupling / 3

            elif type1 == 'pz' and type2 == 'px' and s1 == 0 and s2 == 1:
                return -self.so_coupling / 3

            elif type1 == 'pz' and type2 == 'py' and s1 == 0 and s2 == 1:
                return 1j * self.so_coupling / 3

            elif type1 == 'px' and type2 == 'py' and s1 == 1 and s2 == 1:
                return 1j * self.so_coupling / 3

            elif type1 == 'py' and type2 == 'px' and s1 == 0 and s2 == 0:
                return 1j * self.so_coupling / 3

            elif type1 == 'pz' and type2 == 'px' and s1 == 1 and s2 == 0:
                return self.so_coupling / 3

            elif type1 == 'pz' and type2 == 'py' and s1 == 1 and s2 == 0:
                return 1j * self.so_coupling / 3

            elif type1 == 'px' and type2 == 'pz' and s1 == 1 and s2 == 0:
                return -self.so_coupling / 3

            elif type1 == 'py' and type2 == 'pz' and s1 == 1 and s2 == 0:
                return -1j * self.so_coupling / 3

            elif type1 == 'py' and type2 == 'px' and s1 == 1 and s2 == 1:
                return -1j * self.so_coupling / 3
            else:
                return 0
        else:
            return 0

    def _reset_periodic_bc(self):
        """Reset the matrices determining periodic boundary conditions to their default state
        :return:

        Parameters
        ----------

        Returns
        -------

        """

        self.h_matrix_bc_add = np.zeros((self.basis_size, self.basis_size), dtype=complex)
        self.ov_matrix_bc_add = np.zeros((self.basis_size, self.basis_size), dtype=complex)
        self.h_matrix_bc_factor = np.ones((self.basis_size, self.basis_size), dtype=complex)
        self.k_vector = None

    def _compute_h_matrix_bc_factor(self):
        """Compute the exponential Bloch factors needed when the periodic boundary conditions are applied."""

        for j1 in range(self.num_of_nodes):

            list_of_neighbours = self.get_neighbours(j1)

            for j2 in list_of_neighbours:
                if j1 != j2:
                    coords = np.array(list(self.atom_list.values())[j1], dtype=float) - \
                             np.array(list(self.atom_list.values())[j2], dtype=float)
                    phase = np.exp(1j * np.dot(self.k_vector, coords))

                    for l1 in range(self.orbitals_dict[list(self.atom_list.keys())[j1]].num_of_orbitals):
                        for l2 in range(self.orbitals_dict[list(self.atom_list.keys())[j2]].num_of_orbitals):
                            ind1 = self.qn2ind([('atoms', j1), ('l', l1)], )
                            ind2 = self.qn2ind([('atoms', j2), ('l', l2)], )

                            self.h_matrix_bc_factor[ind1, ind2] = phase
                            # self.h_matrix[ind2, ind1] = self.h_matrix[ind1, ind2]

    def _compute_h_matrix_bc_add(self, split_the_leads=False, overlap=False):
        """Compute additive Bloch exponentials needed to specify pbc

        Parameters
        ----------
        split_the_leads :
             (Default value = False)
        overlap :
             (Default value = False)
        Returns
        -------

        """

        two_leads = False

        if self.ct is not None:
            if np.array(self.ct.pcv).shape[0] == 1:
                two_leads = True

            if split_the_leads:
                if two_leads:
                    flag = None
                else:
                    flag = 'L'

            # loop through all interfacial atoms
            for j1 in self.ct.interfacial_atoms_ind:

                list_of_neighbours = self.ct.get_neighbours(list(self.atom_list.values())[j1])

                for j2 in list_of_neighbours:

                    coords = np.array(list(self.atom_list.values())[j1]) - \
                             np.array(list(self.ct.virtual_and_interfacial_atoms.values())[j2])

                    if split_the_leads and two_leads:
                        flag = self.ct.atom_classifier(list(self.ct.virtual_and_interfacial_atoms.values())[j2],
                                                       self.ct.pcv[0])

                    phase = np.exp(1j * np.dot(self.k_vector, coords))

                    ind = int(list(self.ct.virtual_and_interfacial_atoms.keys())[j2].split('_')[2])

                    for l1 in range(self.orbitals_dict[list(self.atom_list.keys())[j1]].num_of_orbitals):
                        for l2 in range(self.orbitals_dict[list(self.atom_list.keys())[ind]].num_of_orbitals):

                            ind1 = self.qn2ind([('atoms', j1), ('l', l1)])
                            ind2 = self.qn2ind([('atoms', ind), ('l', l2)])

                            if split_the_leads:
                                if flag == 'R':
                                    self.h_matrix_left_lead[ind1, ind2] += phase * \
                                                                           self._get_me(j1, ind, l1, l2, coords=coords)
                                elif flag == 'L':
                                    self.h_matrix_right_lead[ind1, ind2] += phase * \
                                                                            self._get_me(j1, ind, l1, l2, coords=coords)
                                else:
                                    raise ValueError("Wrong flag value")
                            else:
                                self.h_matrix_bc_add[ind1, ind2] += phase * \
                                                                    self._get_me(j1, ind, l1, l2, coords)
                                if overlap:
                                    self.ov_matrix_bc_add[ind1, ind2] += phase * \
                                                                         self._get_me(j1, ind, l1, l2,
                                                                                      coords=coords, overlap=True)

    def get_hamiltonians(self):
        """Return a list of Hamiltonian matrices. For 1D systems, the list is [Hl, Hc, Hr],
        where Hc is the Hamiltonian describing interactions between atoms within a unit cell,
        Hl and Hr are Hamiltonians describing couplings between atoms in the unit cell
        and atoms in the left and right adjacent unit cells.

        Parameters
        ----------

        Returns
        -------
        list
            list of Hamiltonians

        """

        self.k_vector = [0.0, 0.0, 0.0]

        self.h_matrix_left_lead = np.zeros((self.basis_size, self.basis_size), dtype=complex)
        self.h_matrix_right_lead = np.zeros((self.basis_size, self.basis_size), dtype=complex)

        self._compute_h_matrix_bc_add(split_the_leads=True)
        self.k_vector = None

        return self.h_matrix_left_lead.T, self.h_matrix, self.h_matrix_right_lead.T

    def get_site_coordinates(self):
        """Return coordinates of atoms.

        Parameters
        ----------

        Returns
        -------
        numpy.ndarray
            atomic coordinates

        """

        return np.array(self._coords)

    def get_hamiltonians_block_tridiagonal(self, left=-1, right=-1, optimized=True):
        """

        Parameters
        ----------
        left :
             (Default value = None)
             (Default value = None)
        right :
             (Default value = None)
        optimized :
             (Default value = True)

        Returns
        -------

        """

        hl, h0, hr = self.get_hamiltonians()

        if left == -1 and right == -1:
            h_r_h = find_nonzero_lines(hr, 'bottom')
            h_r_v = find_nonzero_lines(hr[-h_r_h:, :], 'left')
            h_l_h = find_nonzero_lines(hl, 'top')
            h_l_v = find_nonzero_lines(hl[:h_l_h, :], 'right')
            left = max(h_l_h, h_r_v)
            right = max(h_r_h, h_l_v)

        if optimized:
            subblocks = split_into_subblocks_optimized(h0, left=left, right=right)
        else:
            subblocks = split_into_subblocks(h0, left, right)

        h01, hl1, hr1 = cut_in_blocks(h0, subblocks)

        if left is not None and right is not None:
            hl1.append(hl[:left, -right:])
            hr1.append(hr[-right:, :left])

        return hl1, h01, hr1, subblocks



import numpy as np
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt


def plot_atom_positions(atom_list, virtual_and_interfacial_atoms, radial_dep):
    """

    Parameters
    ----------
    atom_list :
        
    virtual_and_interfacial_atoms :
        
    radial_dep :
        

    Returns
    -------

    """

    # from mpl_toolkits.mplot3d import Axes3D
    # import matplotlib.pyplot as plt
    # fig = plt.figure()
    # ax = fig.add_subplot(111, projection='3d')
    # coordinates_to_plot = np.asarray(list(self.atom_list.values()))
    # ax.scatter(coordinates_to_plot[:, 0], coordinates_to_plot[:, 1], coordinates_to_plot[:, 2], c='red', s=100)
    #
    # map1 = [item.startswith('*_') for item in list(self.ct.virtual_and_interfacial_atoms.keys())]
    # map2 = [item.startswith('**_') for item in list(self.ct.virtual_and_interfacial_atoms.keys())]
    #
    # coordinates_to_plot = np.asarray(list(self.ct.virtual_and_interfacial_atoms.values()))
    # ax.scatter(coordinates_to_plot[map1, 0], coordinates_to_plot[map1, 1], coordinates_to_plot[map1, 2],
    #            c='green', s=70)
    # ax.scatter(coordinates_to_plot[map2, 0], coordinates_to_plot[map2, 1], coordinates_to_plot[map2, 2],
    #            s=20)
    #
    # ax.set_xlim(-6, 6)
    # ax.set_ylim(-6, 6)
    # ax.set_zlim(-6, 6)
    # plt.show()

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    coordinates_of_atoms_in_unit_cell = np.asarray(list(atom_list.values()))
    ax.scatter(coordinates_of_atoms_in_unit_cell[:, 0], coordinates_of_atoms_in_unit_cell[:, 1], coordinates_of_atoms_in_unit_cell[:, 2], c='k', s=60)

    coordinates_of_atoms_outside_of_unit_cell = np.asarray(list(virtual_and_interfacial_atoms.values()))
    for ii in range(len(atom_list)):
        for jj in range(len(virtual_and_interfacial_atoms)):
            if radial_dep(coordinates_of_atoms_outside_of_unit_cell[jj] - coordinates_of_atoms_in_unit_cell[ii]) == 1:
                ax.scatter(coordinates_of_atoms_outside_of_unit_cell[jj, 0],
                           coordinates_of_atoms_outside_of_unit_cell[jj, 1],
                           coordinates_of_atoms_outside_of_unit_cell[jj, 2],
                           c='r', s=30, alpha=0.3)
            elif radial_dep(coordinates_of_atoms_outside_of_unit_cell[jj] - coordinates_of_atoms_in_unit_cell[ii]) == 2:
                ax.scatter(coordinates_of_atoms_outside_of_unit_cell[jj, 0],
                           coordinates_of_atoms_outside_of_unit_cell[jj, 1],
                           coordinates_of_atoms_outside_of_unit_cell[jj, 2],
                           c='g', s=20, alpha=0.3)
            elif radial_dep(coordinates_of_atoms_outside_of_unit_cell[jj] - coordinates_of_atoms_in_unit_cell[ii]) == 3:
                ax.scatter(coordinates_of_atoms_outside_of_unit_cell[jj, 0],
                           coordinates_of_atoms_outside_of_unit_cell[jj, 1],
                           coordinates_of_atoms_outside_of_unit_cell[jj, 2],
                           c='b', s=10, alpha=1.0)

    ax.set_xlim(-6, 6)
    ax.set_ylim(-6, 6)
    ax.set_zlim(-6, 6)
    plt.show()


def plot_bs_split(kk, vba, cba):
    """

    Parameters
    ----------
    kk :
        
    vba :
        
    cba :
        

    Returns
    -------

    """

    fig, ax = plt.subplots(1, 2)
    ax[0].set_ylim(-1.0, -0.3)
    ax[0].plot(kk, np.sort(np.real(vba)), 'k')
    ax[0].set_xlabel(r'Wave vector ($\frac{\pi}{a}$)')
    ax[0].set_ylabel(r'Energy (eV)')
    ax[0].set_title('Valence band')
    # plt.savefig('bs_vb.pdf')

    ax[1].set_ylim(2.0, 2.7)
    ax[1].plot(kk, np.sort(np.real(cba)), 'k')
    ax[1].set_xlabel(r'Wave vector ($\frac{\pi}{a}$)')
    ax[1].set_ylabel(r'Energy (eV)')
    ax[1].set_title('Conduction band')
    fig.tight_layout()
    # plt.savefig('bs_cb.pdf')
    plt.show()


def plot_atom_positions1(h, virtual_and_interfacial_atoms, radial_dep):
    """

    Parameters
    ----------
    h :
        
    virtual_and_interfacial_atoms :
        
    radial_dep :
        

    Returns
    -------

    """

    # from mpl_toolkits.mplot3d import Axes3D
    # import matplotlib.pyplot as plt
    # fig = plt.figure()
    # ax = fig.add_subplot(111, projection='3d')
    # coordinates_to_plot = np.asarray(list(self.atom_list.values()))
    # ax.scatter(coordinates_to_plot[:, 0], coordinates_to_plot[:, 1], coordinates_to_plot[:, 2], c='red', s=100)
    #
    # map1 = [item.startswith('*_') for item in list(self.ct.virtual_and_interfacial_atoms.keys())]
    # map2 = [item.startswith('**_') for item in list(self.ct.virtual_and_interfacial_atoms.keys())]
    #
    # coordinates_to_plot = np.asarray(list(self.ct.virtual_and_interfacial_atoms.values()))
    # ax.scatter(coordinates_to_plot[map1, 0], coordinates_to_plot[map1, 1], coordinates_to_plot[map1, 2],
    #            c='green', s=70)
    # ax.scatter(coordinates_to_plot[map2, 0], coordinates_to_plot[map2, 1], coordinates_to_plot[map2, 2],
    #            s=20)
    #
    # ax.set_xlim(-6, 6)
    # ax.set_ylim(-6, 6)
    # ax.set_zlim(-6, 6)
    # plt.show()

    atom_list = h.atom_list
    virtual_and_interfacial_atoms = h.ct.virtual_and_interfacial_atoms

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    neigbours_in_unit_cell = [h.get_neighbours(coord) for coord in list(atom_list.values())]
    coordinates_of_atoms_in_unit_cell = [list(atom_list.values())[item] for item in np.array(neigbours_in_unit_cell).flatten()]
    coordinates_of_atoms_in_unit_cell = np.array(coordinates_of_atoms_in_unit_cell)
    ax.scatter(coordinates_of_atoms_in_unit_cell[:, 0], coordinates_of_atoms_in_unit_cell[:, 1], coordinates_of_atoms_in_unit_cell[:, 2], c='k', s=60)

    neigbours_out_unit_cell = [h.ct.get_neighbours(coord) for coord in list(atom_list.values())]

    coordinates_of_atoms_out_unit_cell1 = [list(virtual_and_interfacial_atoms.values())[item] for item in
                                          np.array(neigbours_out_unit_cell)[0]]

    coordinates_of_atoms_out_unit_cell2 = [list(virtual_and_interfacial_atoms.values())[item] for item in
                                          np.array(neigbours_out_unit_cell)[1]]
    coordinates_of_atoms_outside_of_unit_cell1 = np.array(coordinates_of_atoms_out_unit_cell1)
    coordinates_of_atoms_outside_of_unit_cell2 = np.array(coordinates_of_atoms_out_unit_cell2)
    # coordinates_of_atoms_outside_of_unit_cell = np.asarray(list(virtual_and_interfacial_atoms.values()))
    ax.scatter(coordinates_of_atoms_outside_of_unit_cell1[:, 0], coordinates_of_atoms_outside_of_unit_cell1[:, 1],
               coordinates_of_atoms_outside_of_unit_cell1[:, 2], c='r', s=30)
    ax.scatter(coordinates_of_atoms_outside_of_unit_cell2[:, 0], coordinates_of_atoms_outside_of_unit_cell2[:, 1],
               coordinates_of_atoms_outside_of_unit_cell2[:, 2], c='b', s=60)

    ax.set_xlim(-6, 6)
    ax.set_ylim(-6, 6)
    ax.set_zlim(-6, 6)
    plt.show()






import unittest
import numpy as np
from nanonet.tb.aux_functions import compute_edge, blocksandborders_constrained


class Test(unittest.TestCase):
    """ """

    def setUp(self):
        """ """
        a = np.array([[1, 1], [1, 1]])
        b = np.zeros((2, 2))
        self.test_matrix = np.block([[a, b, b], [b, a, b], [b, b, a]])
        self.test_matrix[0, 2] = 2
        self.test_matrix[2, 0] = 2
        self.test_matrix[2, 4] = 3
        self.test_matrix[4, 2] = 3

        self.edge = np.array([3, 3, 5, 5, 6, 6])
        self.edge_star = np.array([2, 4, 4, 6, 6, 6])

    def test_compute_edge(self):
        """ """

        edge, edge1 = compute_edge(self.test_matrix)
        np.testing.assert_allclose(edge, self.edge, atol=1e-4)
        np.testing.assert_allclose(edge1, self.edge_star, atol=1e-4)

    def test_blocksandborders_constrained_0_0(self):
        """ """
        ans = blocksandborders_constrained(0, 0, self.edge, self.edge_star)
        self.assertListEqual(ans, [1, 3, 1, 1])

    def test_blocksandborders_constrained_1_1(self):
        """ """
        ans = blocksandborders_constrained(1, 1, self.edge, self.edge_star)
        self.assertListEqual(ans, [1, 3, 1, 1])

    def test_blocksandborders_constrained_2_2(self):
        """ """
        ans = blocksandborders_constrained(2, 2, self.edge, self.edge_star)
        self.assertListEqual(ans, [2, 2, 2])

    def test_blocksandborders_constrained_3_3(self):
        """ """
        ans = blocksandborders_constrained(3, 3, self.edge, self.edge_star)
        self.assertListEqual(ans, [3, 3])

    def test_blocksandborders_constrained_4_4(self):
        """ """
        ans = blocksandborders_constrained(4, 4, self.edge, self.edge_star)
        self.assertListEqual(ans, [6])

    def test_blocksandborders_constrained_5_5(self):
        """ """
        ans = blocksandborders_constrained(6, 6, self.edge, self.edge_star)
        self.assertListEqual(ans, [6])

    def test_blocksandborders_constrained_2_4(self):
        """ """
        ans = blocksandborders_constrained(2, 4, self.edge, self.edge_star)
        self.assertListEqual(ans, [2, 4])

    def test_blocksandborders_constrained_2_3(self):
        """ """
        ans = blocksandborders_constrained(2, 3, self.edge, self.edge_star)
        self.assertListEqual(ans, [2, 1, 3])

    def test_blocksandborders_constrained_3_2(self):
        """ """
        ans = blocksandborders_constrained(3, 2, self.edge, self.edge_star)
        self.assertListEqual(ans, [4, 2])


if __name__ == "__main__":
    unittest.main()



import numpy as np
from scipy.linalg import eig, eigh_tridiagonal
from nanonet.negf.hamiltonian_chain import fd
import time



def t_order_frac(x):
    """

    Parameters
    ----------
    x :
        

    Returns
    -------

    """
    return 0.5 * (np.sign(x) + 1.0) / x


def approximant(energy, poles, residues):
    """

    Parameters
    ----------
    energy :
        
    poles :
        
    residues :
        

    Returns
    -------

    """

    arg = np.array(energy)
    ans = np.zeros(arg.shape) + 0.5

    for j in range(len(poles)):
        if poles[j] > 0:
            ans = ans - residues[j] / (arg - 1j / poles[j]) - residues[j] / (arg + 1j / poles[j])

    return ans


def approximant_diff(energy, poles, residues):
    """

    Parameters
    ----------
    energy :
        
    poles :
        
    residues :
        

    Returns
    -------

    """

    arg = np.array(energy)
    ans = np.zeros(arg.shape) + 0.5 * 0

    for j in range(len(poles)):
        if poles[j] > 0:
            ans = ans - residues[j] / (arg - 1j / poles[j]) ** 2 - residues[j] / (arg + 1j / poles[j]) ** 2

    return ans


def poles_and_residues(cutoff=2):
    """

    Parameters
    ----------
    cutoff :
         (Default value = 2)

    Returns
    -------

    """

    # b_mat = [1 / (2.0 * np.sqrt((2*(j+1) - 1)*(2*(j+1) + 1))) for j in range(0, cutoff-1)]

    j = np.arange(cutoff-1)
    b_mat = 1 / (2.0 * np.sqrt((2*(j+1) - 1)*(2*(j+1) + 1)))
    # b_mat = np.diag(b_mat, -1) + np.diag(b_mat, 1)
    # poles, residues = eig(b_mat)
    poles, residues = eigh_tridiagonal(np.zeros((len(b_mat) + 1,)), b_mat)

    # arg = np.argmax(np.abs(residues), axis=0)

    residues = 0.25 * np.array([np.abs(residues[0, j])**2 / (poles[j] ** 2) for j in range(residues.shape[0])])

    return poles, residues


def test_gf1(z):
    """

    Parameters
    ----------
    z :
        

    Returns
    -------

    """
    return t_order_frac(z + 10.0) + \
           t_order_frac(z + 5.0) + \
           t_order_frac(z + 2.0) + \
           t_order_frac(z - 7.0)


def test_gf(z):
    """

    Parameters
    ----------
    z :
        

    Returns
    -------

    """
    return 1.0 / (z + 10.0) + \
           1.0 / (z + 5.0) + \
           1.0 / (z + 2.0) + \
           1.0 / (z - 7.0)


def test_integration(Ef, tempr, cutoff=70, gf=test_gf):
    """

    Parameters
    ----------
    Ef :
        
    tempr :
        
    cutoff :
         (Default value = 70)
    gf :
         (Default value = test_gf)

    Returns
    -------

    """

    R = 1.0e20

    a1, b1 = poles_and_residues(cutoff=cutoff)
    zero_moment = 1j * R * gf(1j * R)

    ans = 0
    # Ef = -5

    betha = 1.0 / (8.617333262145e-5 * tempr)

    b1 = b1[np.real(a1) > 0]
    a1 = a1[np.real(a1) > 0]

    aaa = Ef + 1j / a1 / betha
    ans1 = np.sum(4 * 1j / betha * gf(aaa) * b1)

    # for j in range(len(a1)):
    #     if np.real(a1[j]) > 0:
    #         aaa = Ef + 1j / a1[j] / betha
    #         ans = ans + 4 * 1j / betha * gf(aaa) * b1[j]
    #         # print(np.imag(ans), a1[j], b1[j])

    return -zero_moment + np.imag(ans1)


def bf_integration(Ef, tempr, gf=test_gf):
    """

    Parameters
    ----------
    Ef :
        
    tempr :
        
    gf :
         (Default value = test_gf)

    Returns
    -------

    """

    # temp = 100
    R = 2e2
    # Ef = 2.1

    energy = np.linspace(-R+Ef, R+Ef, int(3e5))
    ans = np.imag(np.trapz(test_gf(energy + 1j * 10e-4) * fd(energy, Ef, tempr), energy))

    return -1.0 / np.pi * ans


if __name__=='__main__':

    import matplotlib.pyplot as plt
    # ans = bf_integration(gf=test_gf)

    Ef = np.linspace(-70, 70, 150)
    ans = []
    ans1 = []

    t = time.process_time()
    for ef in Ef:
        ans1.append(test_integration(ef, 10, cutoff=500, gf=test_gf1))
    print(time.process_time() - t)

    t = time.process_time()
    for ef in Ef:
        ans.append(bf_integration(ef, 10, gf=test_gf))
    print(time.process_time() - t)

    plt.plot(Ef, np.array(ans))
    plt.plot(Ef, np.array(ans1), 'o-')
    plt.show()
    print(ans, ans1)

    a1, b1 = poles_and_residues(cutoff=2)
    a2, b2 = poles_and_residues(cutoff=10)
    a3, b3 = poles_and_residues(cutoff=30)
    a4, b4 = poles_and_residues(cutoff=50)
    a5, b5 = poles_and_residues(cutoff=100)

    print(a5, b5)

    energy = np.linspace(-5.7, 5.7, 3000)

    temp = 300
    fd0 = fd(energy, 0, temp)

    kb = 8.61733e-5  # Boltzmann constant in eV
    energy = energy / (kb * temp)

    ans1 = approximant(energy, a1, b1)
    ans2 = approximant(energy, a2, b2)
    ans3 = approximant(energy, a3, b3)
    ans4 = approximant(energy, a4, b4)
    ans5 = approximant(energy, a5, b5)

    plt.plot(energy, fd0)
    plt.plot(energy, ans1)
    plt.plot(energy, ans2)
    plt.plot(energy, ans3)
    plt.plot(energy, ans4)
    plt.plot(energy, ans5)
    plt.show()
    #



import numpy as np
import scipy.linalg as linalg
import scipy.optimize as opt
import cmath


def pole_maker(Emin, ChemPot, kT, reltol):
    """This is an alternate pole summation method implemented by Areshkin-Nikolic [CITE].
    Similar to the continued-fraction representation of Ozaki, this method allows for
    efficient computation of the density matrix by complex pole summation. Both methods
    approximate the Fermi-Dirac distribution function so that it is more easily manipulable.

    One of the exciting uses of this method is for finite differences, and the computation
    of quasi-equilibrium solutions for use in efficient non-equilibrium density matrices
    as the user can reuse certain poles but with different weights for little extra cost.

    What A-N does is replaces f(E,mu) with one that approximates it on the real line:
    f(E,mu,kT) -> f(E,1j*mu_im,1j*kT_im)*( f(E,mu,kT) - f(Emin, mu_re,kT_re) ) but has
    drastic changes in the upper complex plane. By judicious choice of mu_im to be at
    least p*kT away from the real line and mu_re to be at least p*kT away from the
    minimum eigenvalue Emin, this gives a controlled relative error of e^-p.

    This function automatically switches from first order to second order when
    appropriate. This is when the interval divided by the temperature exceeds 10^3.
    To-do: add third order method.

     Parameters
    ----------
    Emin    : scalar (dtype=numpy.float)
           Minimum occupied state, e.g. a band edge.
    ChemPot : scalar (dtype=numpy.float)
           The chemical potential
    kT      : scalar (dtype=numpy.float)
           The temperature (in units of energy)
    reltol : scalar (dtype=numpy.float)
           The desired relative tolerance. p = -np.log(reltol)

    """
    p = -np.log(reltol)  # Compute the exponent for the relative tolerance desired.

    # When energy exceeds 10^3, switch to second order poles.
    z = (ChemPot - Emin) / kT

    if z < 10 ** 3:
        poles, residues = pole_order_one(Emin, ChemPot, kT, p)
    else:
        poles, residues = pole_order_two(Emin, ChemPot, kT, p)

    return poles, residues


def pole_order_one(Emin, ChemPot, kT, p):
    kTRe, kTIm, muRe, muIm = pole_minimizer_one(Emin, ChemPot, kT, p)

    # Three pole branches
    #   :  :
    # - - - - C
    #   :  :
    #   :  :
    #   A  B

    dummyindex = np.arange(-10000, 10000 + 1)  # Integers from -500 to 500 inclusive,.
    # Need to make a more reasonable guess for valid integers, real line is easy, imag is hard

    # Residues for Fermi-Dirac are simply -kT, residue theorem multiplies
    # by 2*pi*i, then the windowing function multiplies in its own factor:

    # Branch A poles:
    poleA = muRe + 1j * np.pi * kTRe * (2 * dummyindex + 1)  # Analytical pole location from definition of fermi-fun
    # Compute their residues
    resA = (2j * np.pi * kTRe) * fermi_fun(poleA, 1j * muIm, 1j * kTIm)
    # Determine which to trim, they must have magnitude over tol and be in upper complex plane
    zA = (np.abs(resA) >= np.exp(-p) * kT) & (np.imag(poleA) > 0)
    poleA = poleA[zA]
    resA = resA[zA]

    # Branch B poles:
    poleB = ChemPot + 1j * np.pi * kT * (2 * dummyindex + 1)  # Analytical pole location from definition of fermi-fun
    # Compute their residues
    resB = -(2j * np.pi * kT) * fermi_fun(poleB, 1j * muIm, 1j * kTIm)
    # Determine which to trim, they must have magnitude over tol and be in upper complex plane
    zB = (np.abs(resB) >= np.exp(-p) * kT) & (np.imag(poleB) > 0)
    poleB = poleB[zB]
    resB = resB[zB]

    # Branch C poles:
    poleC = 1j * muIm - np.pi * kTIm * (2 * dummyindex + 1)  # Analytical pole location from definition of fermi-fun
    # Residues
    resC = (2 * np.pi * kTIm) * (fermi_fun(poleC, ChemPot, kT) - fermi_fun(poleC, muRe, kTRe))

    zC = (np.abs(resC) >= np.exp(-p) * kT) & (np.imag(poleC) > 0)
    poleC = poleC[zC]
    resC = resC[zC]

    poles = np.concatenate((poleA, poleB, poleC))
    residues = np.concatenate((resA, resB, resC))

    return poles, residues


def pole_minimizer_one(Emin, ChemPot, kT, p):
    """
    This function is minimizing the problem:
    N = N_AB + N_BC + N_CD, where:
    N_AB = (muIm + p*kTIm)/(2*pi*kT)
    N_BC = ( (mu + p*kT) - (muRe - p*kTre) )/(2*pi*kTIm)
    N_CD = (muIm + p*kTIm)/(2*pi*kTRe)
    though this has 4 free parameters (excluding p), we
    set certain parameters so that we do not violate
    our relative tolerance, these are:
    muRe = Emin - p*kTre
    muIm = p*kTim
    which then reduces this to a two variable problem:
    N = 2*p*kTIm*(1/kT + 1/kTRe) +
        ( (ChemPot - Emin) + p*(kT + 2*kTRe) )/( 2*kTIm )
    The function then optimizes the position of the
    pole branches not to overlap then outputs the temp-
    eratures and chemical potentials
    """

    # There's a lot going on here, first we desig-
    # nate the cost function for the whole problem
    def pole_cost_one_A(z):
        kTReA = z[0]  # Added the A to stop it from shadowing
        kTImA = z[1]  # the variable later on in the function.
        return 2 * p * kTImA * (1 / kT + 1 / kTReA) + (ChemPot - Emin + p * (kT + 2 * kTReA)) / (2 * kTImA)

    z0 = np.array([kT, kT])
    zAout = opt.minimize(pole_cost_one_A, z0, bounds=((kT, None), (kT, None)), tol=1e-8)

    # Now we modify it so that the imaginary chemical potential is
    # directly in-between the poles generated from the real line
    l_opt = np.ceil(p * zAout.x[1] / (2 * np.pi * kT))
    muIm = np.pi * kT * (2 * l_opt)
    kTIm = muIm / p  # Because muIm = p kTIm, kTIm = muIm/p

    # Now that we have fixed the imaginary part, we re-optimize
    # kTRe, because some marginal improvement may be made after
    # applying the previous constraints

    def pole_cost_one_B(z):
        kTReB = z[0]  # Added the zeros to stop it from shadowing
        return 2 * p * kTIm * (1 / kT + 1 / kTReB) + (ChemPot - Emin + p * (kT + 2 * kTReB)) / (2 * kTIm)

    z1 = np.array(zAout.x[0])
    bndsB = [(kT, None)]
    zBout = opt.minimize(pole_cost_one_B, z1, bounds=bndsB, tol=1e-8)

    # After it is optimized we once again correct the ans-
    # wer to be between the poles of the nearest branch,
    # thereby preventing a second order pole situation.

    m_opt = np.ceil(np.abs(ChemPot - Emin + p * zBout.x[0]) / (2 * np.pi * kTIm))
    muRe = ChemPot - np.pi * kTIm * (2 * m_opt)
    kTRe = (Emin - muRe) / p

    # The extra poles introduced from this
    # positional fixing process is negligible
    # and is typically no greater than a small
    # handful of poles.

    return kTRe, kTIm, muRe, muIm


def pole_order_two(Emin, ChemPot, kT, p):
    kTRe1, kTIm1, muRe1, muIm1, kTRe2, kTIm2, muRe2, muIm2 = pole_minimizer_two(Emin, ChemPot, kT, p)

    poles = 1
    residues = 1

    return poles, residues


def pole_minimizer_two(Emin, ChemPot, kT, p):
    """
    This function is minimizing the problem:
    N = N_AB + N_BC + N_CD + N_DE + N_EF, where:
    N_AB =
    N_BC =
    N_CD =
    N_DE =
    N_EF =
    though this has x free parameters (excluding p), we
    set certain parameters so that we do not violate
    our relative tolerance, these are:
    muRe1 = x
    muIm1 = y
    muRe2 = x
    muIm2 = y
    which then reduces this to a x variable problem:
    N =

    """

    # Run code, get output, real answer will be different than these initial filler values.
    kTIm1 = kT
    kTIm2 = kT
    kTRe1 = kT
    kTRe2 = kT
    muRe2 = Emin - p * kTRe2
    muRe1 = 0.5 * (muRe2 + ChemPot)  # Initial guess is halfway between
    muIm1 = p * kTIm1
    muIm2 = p * kTIm2

    return kTRe1, kTIm1, muRe1, muIm1, kTRe2, kTIm2, muRe2, muIm2


def pole_finite_difference(muL, muR, kT, reltol):
    """
    For computing the finite difference derivative using pole summation, see Vaitkus thesis.

    By using 2x one-sided differences, we can get the forwards, backwards, and centred
    first derivatives and the centred second derivative. The poles will be written as a vector
    containing all points, and the residues will be two vectors containing the poles for the
    backwards and forward differences respectively as (f_mid-f_min) and (f_max-f_mid).

    This choice is so that the user need compute the poles only once and
    can carry out whichever residues you like using the following recipes:

    h = np.abs(muR - muC)

    Forward  1st = ( (f_max-f_mid)                 ) / (   h)
    Centred  1st = ( (f_max-f_mid) + (f_mid-f_min) ) / ( 2*h)
    Backward 1st = ( (f_mid-f_min)                 ) / (   h)
    Centred  2nd = ( (f_max-f_mid) - (f_mid-f_min) ) / (h**2)

    Parameters
    ----------
    muL    : scalar (dtype=numpy.float)
           Left chemical potential
    muR    : scalar (dtype=numpy.float)
           Right chemical potential
    kT     : scalar (dtype=numpy.float)
           The temperature (in units of energy)
    reltol : scalar (dtype=numpy.float)
           The desired relative tolerance. p = -np.log(reltol)

    """

    p = -np.log(reltol)  # Value of p that gives desired relative tolerance.

    muMin = np.min([muL, muR])
    muMax = np.max([muL, muR])
    muMid = 0.5 * (muMin + muMax)

    kTIm = np.sqrt((muMax - muMin + 2 * p * kT) / (6 * p / kT))  # Analytical solution for minimum pole number
    # Correcting temperature so that p*kTIm will be directly between the poles from the other funcs.
    kTIm = (2 * np.pi * kT / p) * np.ceil((p * kTIm) / (2 * np.pi * kT))
    muIm = p * kTIm  # Where muIm should be to get e^-p error

    # Four pole branches
    #   :  :  :
    # - - - - - - D
    #   :  :  :
    #   :  :  :
    #   A  B  C

    dummyindex = np.arange(-10000, 10000 + 1)  # Integers from -500 to 500 inclusive,.
    # Need to make a more reasonable guess for valid integers, real line is easy, imag is hard

    # Residues for Fermi-Dirac are simply -kT, residue theorem multiplies
    # by 2*pi*i, then the windowing function multiplies in its own factor:

    # Branch A poles:
    poleA = muMin + 1j * np.pi * kT * (2 * dummyindex + 1)  # Analytical pole location from definition of fermi-fun
    # Residues for left
    resA_L = (2j * np.pi * kT) * fermi_fun(poleA, 1j * muIm, 1j * kTIm)
    # Residues for right
    resA_R = 0 * poleA
    # Determine which to trim, they must have magnitude over tol and be in upper complex plane
    zA = (np.abs(resA_L) >= np.exp(-p) * kT) & (np.imag(poleA) > 0)
    poleA = poleA[zA]
    resA_L = resA_L[zA]
    resA_R = resA_R[zA]

    # Branch B poles:
    poleB = muMid + 1j * np.pi * kT * (2 * dummyindex + 1)  # Analytical pole location from definition of fermi-fun
    # Residues for left
    resB_L = -(2j * np.pi * kT) * fermi_fun(poleB, 1j * muIm, 1j * kTIm)
    # Residues for right are exactly the negatives of the left by design
    resB_R = -resB_L
    # Determine which to trim
    zB = (np.abs(resB_L) >= np.exp(-p) * kT) & (np.imag(poleB) > 0)
    poleB = poleB[zB]
    resB_L = resB_L[zB]
    resB_R = resB_R[zB]

    # Branch C poles:
    poleC = muMax + 1j * np.pi * kT * (2 * dummyindex + 1)  # Analytical pole location from definition of fermi-fun
    # Residues for left
    resC_L = 0 * poleC
    # Residues for right
    resC_R = -(2j * np.pi * kT) * fermi_fun(poleC, 1j * muIm, 1j * kTIm)
    # Determine which to trim
    zC = (np.abs(resC_R) >= np.exp(-p) * kT) & (np.imag(poleC) > 0)
    poleC = poleC[zC]
    resC_L = resC_L[zC]
    resC_R = resC_R[zC]

    # Branch D poles:
    poleD = 1j * muIm - np.pi * kTIm * (2 * dummyindex + 1)  # Analytical pole location from definition of fermi-fun
    # Residues for left
    resD_L = (2 * np.pi * kTIm) * (fermi_fun(poleD, muMid, kT) - fermi_fun(poleD, muMin, kT))
    # Residues for right
    resD_R = (2 * np.pi * kTIm) * (fermi_fun(poleD, muMax, kT) - fermi_fun(poleD, muMid, kT))
    # Determine which to trim

    zD = np.maximum(np.abs(resD_L), np.abs(resD_R)) > np.exp(-p) * kT
    poleD = poleD[zD]
    resD_L = resD_L[zD]
    resD_R = resD_R[zD]

    poles = np.concatenate((poleA, poleB, poleC, poleD))
    residuesL = np.concatenate((resA_L, resB_L, resC_L, resD_L))
    residuesR = np.concatenate((resA_R, resB_R, resC_R, resD_R))

    return poles, residuesL, residuesR


def fermi_fun(E, mu, kT):
    """
    Computes the Fermi-Dirac distribution function. Auxiliary function just to tidy the workspace.

    Parameters
    ----------
    E    : scalar (dtype=numpy.float)
         Energy being evaluated
    mu   : scalar (dtype=numpy.float)
         Chemical potential
    kT   : scalar (dtype=numpy.float)
         Temperature (in units of energy)
    """
    # Usually the good way
    #    x = (E - mu) / (2 * kT)
    #    return 0.5 * (1 - np.tanh(x))
    # Usual way that most do
    #    x = (E - mu)/kT
    #    return 1/(np.exp(x) + 1)
    # However both cmath/numpy throw errors for either
    # so we fix it using the following:

    x = (E - mu) / kT
    # the function is periodic modulo 2pi, so we subtract
    # off a large integer imaginary part because the math
    # packages numpy/cmath do not like large imaginary parts.
    # Right now it's commented out because of debugging.
    # x = x - 2j * np.pi * np.round(0.5 * np.imag(x) / np.pi)
    out = []

    # stack exchange advice, to avoid overflow issues, make
    # the real part of numbers always smaller than 1 by using
    # alternative definitions for positive and negative Re(x)

    if x.size < 2:
        if np.real(x) <= 0:
            out = 1 / (cmath.exp(x) + 1)
        else:
            out = cmath.exp(-x) / (1 + cmath.exp(-x))
    else:
        for xi in x:
            if np.real(xi) <= 0:
                out.append(1 / (cmath.exp(xi) + 1))
            else:
                out.append(cmath.exp(-xi) / (1 + cmath.exp(-xi)))

    return np.array(out)


def fermi_deriv(E, mu, kT):
    """
    Computes the Fermi-Dirac distribution function derivative. Auxiliary function just to tidy the workspace.

    Parameters
    ----------
    E    : scalar (dtype=numpy.float)
         Energy being evaluated
    mu   : scalar (dtype=numpy.float)
         Chemical potential
    kT   : scalar (dtype=numpy.float)
         Temperature (in units of energy)
    """
    # Tanh should be more numerically stable than 1/(exp(x) + 1), however
    # numpy has issues with large complex values in both exp and tanh
    # x = (E - mu)/kT
    # return 1/(np.exp(x) + 1)

    x = (E - mu) / (2 * kT)
    return (np.cosh(x) ** -2) / (4 * kT)


def fermi_deriv2(E, mu, kT):
    """
    Computes the Fermi-Dirac distribution function second derivative. Auxiliary function just to tidy the workspace.

    Parameters
    ----------
    E    : scalar (dtype=numpy.float)
         Energy being evaluated
    mu   : scalar (dtype=numpy.float)
         Chemical potential
    kT   : scalar (dtype=numpy.float)
         Temperature (in units of energy)
    """
    # Tanh should be more numerically stable than 1/(exp(x) + 1), however
    # numpy has issues with large complex values in both exp and tanh
    # x = (E - mu)/kT
    # return 1/(np.exp(x) + 1)

    x = (E - mu) / (2 * kT)
    return np.tanh(x) * (np.cosh(x) ** -2) / (4 * kT ** 2)



"""
The module contains functions that computes Green's functions and their poles
"""
from __future__ import print_function, division
import numpy as np
import scipy.sparse as sp
import scipy.linalg as linalg


def surface_greens_function_poles(hl, h0, hr):
    """Computes eigenvalues and eigenvectors for the complex band structure problem.
    The eigenvalues correspond to the wave vectors as `exp(ik)`.

    Parameters
    ----------
    hl : numpy.ndarray (dtype=numpy.float)
        Left-side coupling Hamiltonian
    h0 : numpy.ndarray (dtype=numpy.float)
        Hamiltonian of the device region
    hr : numpy.ndarray (dtype=numpy.float)
        Right-side coupling Hamiltonian

    Returns
    -------
    eigenvals : numpy.ndarray (dtype=numpy.complex)
        Array of sorted eigenvalues
    eigenvects : numpy.ndarray (dtype=numpy.complex)
        Eigenvector matrix
    """

    # linearize polynomial eigenvalue problem

    shapes_d = []
    shapes_ud = []
    shapes_ld = []

    if isinstance(h0, list):
        num_blocks = len(h0)
        for j in range(len(h0)):
            shapes_d.append(h0[j].shape[0])
            shapes_ld.append(hl[j].shape)
            shapes_ud.append(hr[j].shape)
        matix_size = np.sum(shapes_d)
    else:
        num_blocks = 1
        shapes_d.append(h0.shape[0])
        shapes_ld.append(hl.shape)
        shapes_ud.append(hr.shape)
        matix_size = h0.shape[0]
        hl = [hl]
        h0 = [h0]
        hr = [hr]

    full_matrix_size = 2 * matix_size
    identity = np.identity(matix_size)
    main_matrix = np.zeros((full_matrix_size, full_matrix_size), dtype=complex)
    overlap_matrix = np.zeros((full_matrix_size, full_matrix_size), dtype=complex)
    main_matrix[0:matix_size, matix_size:2 * matix_size] = identity
    overlap_matrix[0:matix_size, 0:matix_size] = identity
    main_matrix[matix_size:matix_size+hl[-1].shape[0], matix_size-hl[-1].shape[1]:matix_size] = -hl[-1]
    overlap_matrix[2*matix_size-hr[-1].shape[0]:2*matix_size, matix_size:matix_size+hr[-1].shape[1]] = hr[-1]

    offfset = matix_size

    for j in range(num_blocks):

        main_matrix[offfset:offfset+h0[j].shape[0], offfset:offfset+h0[j].shape[0]] = -h0[j]

        if j > 0:
            main_matrix[offfset:offfset + hl[j - 1].shape[0], offfset - hl[j - 1].shape[1]:offfset] = hl[j - 1]
            main_matrix[offfset - hl[j - 1].shape[1]:offfset, offfset:offfset + hl[j - 1].shape[0]] = hr[j - 1]

        offfset += h0[j].shape[0]

    alpha, betha, _, eigenvects, _, _ = linalg.lapack.cggev(main_matrix, overlap_matrix)

    betha[betha == 0] = 1e-19
    eigenvals = alpha / betha

    # sort absolute values
    ind = np.argsort(np.abs(eigenvals))
    eigenvals = eigenvals[ind]
    eigenvects = eigenvects[:, ind]
    eigenvects = eigenvects[matix_size:, :]

    return eigenvals, eigenvects


def group_velocity(eigenvector, eigenvalue, h_r):
    """Computes a group velocity of wave packets

    Parameters
    ----------
    eigenvector : numpy.ndarray (dtype=numpy.complex)
        Eigenvectors
    eigenvalue : numpy.ndarray (dtype=numpy.complex)
        Eigenvalues
    h_r : numpy.matrix numpy.ndarray (dtype=numpy.float)
        Coupling Hamiltonian

    Returns
    -------
    gv : float
        Group velocity of the wave package
    """

    num_vecs = eigenvector.shape[1]
    if num_vecs > 1:
        gv = np.zeros(num_vecs)
        for j in range(num_vecs):
            gv[j] = np.imag(np.dot(np.dot(np.dot(eigenvector[:, j].conj().T, h_r), eigenvalue[j]), eigenvector[:, j]))
    else:
        gv = np.imag(np.dot(np.dot(np.dot(eigenvector.conj().T, h_r), eigenvalue), eigenvector))

    return gv


def iterate_gf(E, h_0, h_l, h_r, se, num_iter):
    """

    Parameters
    ----------
    E : numpy.ndarray (dtype=numpy.float)
        Energy array
    h_l : numpy.ndarray (dtype=numpy.float)
        Left-side coupling Hamiltonian
    h_0 : numpy.ndarray (dtype=numpy.float)
        Hamiltonian of the device region
    h_r : numpy.ndarray (dtype=numpy.float)
        Right-side coupling Hamiltonian
    se : numpy.ndarray (dtype=numpy.complex)
        Self-energy
    num_iter : int
        Number of iterations

    Returns
    -------
    se : numpy.ndarray (dtype=numpy.complex)
        Self-energy
    """

    for _ in range(num_iter):
        se = h_r.dot(linalg.pinv(E * np.identity(h_0.shape[0]) - h_0 - se)).dot(h_l)
    return se


def surface_greens_function(E, h_l, h_0, h_r, iterate=False, damp=0.0001j):
    """Computes surface self-energies using the eigenvalue decomposition.
    The procedure is described in
    [M. Wimmer, Quantum transport in nanostructures: From computational concepts
    to spintronics in graphene and magnetic tunnel junctions, 2009, ISBN-9783868450255].

    Parameters
    ----------
    E : numpy.ndarray (dtype=numpy.float)
        Energy array
    h_l : numpy.ndarray (dtype=numpy.float)
        Left-side coupling Hamiltonian
    h_0 : numpy.ndarray (dtype=numpy.float)
        Hamiltonian of the device region
    h_r : numpy.ndarray (dtype=numpy.float)
        Right-side coupling Hamiltonian
    iterate : bool
        Iterate to stabilize TB matrix (Default value = False)
    damp :
        damping (Default value = 0.0001j)

    Returns
    -------
    sgf_l : numpy.ndarray (dtype=numpy.complex)
        Left-lead self-energy matrix
    sgf_r : numpy.ndarray (dtype=numpy.complex)
        Right-lead self-energy matrix
    """

    # linearize polynomial eigenvalue problem

    shapes_d = []
    shapes_ud = []
    shapes_ld = []

    if isinstance(h_0, list):
        num_blocks = len(h_0)
        for j in range(len(h_0)):
            shapes_d.append(h_0[j].shape[0])
            shapes_ld.append(h_l[j].shape)
            shapes_ud.append(h_r[j].shape)
        matrix_size = np.sum(shapes_d)
    else:
        num_blocks = 1
        shapes_d.append(h_0.shape[0])
        shapes_ld.append(h_l.shape)
        shapes_ud.append(h_r.shape)
        matrix_size = h_0.shape[0]
        h_l = [h_l]
        h_0 = [h_0]
        h_r = [h_r]

    full_matrix_size = 2 * matrix_size
    identity = np.identity(matrix_size)
    main_matrix = np.zeros((full_matrix_size, full_matrix_size), dtype=complex)
    overlap_matrix = np.zeros((full_matrix_size, full_matrix_size), dtype=complex)
    main_matrix[0:matrix_size, matrix_size:2 * matrix_size] = identity
    overlap_matrix[0:matrix_size, 0:matrix_size] = identity
    main_matrix[matrix_size:matrix_size+h_l[-1].shape[0], matrix_size-h_l[-1].shape[1]:matrix_size] = -h_l[-1]
    overlap_matrix[2*matrix_size-h_r[-1].shape[0]:2*matrix_size, matrix_size:matrix_size+h_r[-1].shape[1]] = h_r[-1]

    offset = matrix_size

    for j in range(num_blocks):

        main_matrix[offset:offset+h_0[j].shape[0], offset:offset+h_0[j].shape[0]] = -h_0[j]

        if j > 0:
            main_matrix[offset:offset + h_l[j - 1].shape[0], offset - h_l[j - 1].shape[1]:offset] = -h_l[j - 1]
            main_matrix[offset - h_l[j - 1].shape[1]:offset, offset:offset + h_l[j - 1].shape[0]] = -h_r[j - 1]

        offset += h_0[j].shape[0]

    main_matrix[matrix_size:2*matrix_size, matrix_size:2*matrix_size] = \
        (E + damp) * np.identity(matrix_size) + main_matrix[matrix_size:2*matrix_size, matrix_size:2*matrix_size]

    alpha, betha, _, vects, _, _ = linalg.lapack.cggev(main_matrix, overlap_matrix)

    betha[betha == 0] = 1e-19
    vals = alpha / betha

    # sort absolute values
    ind = np.argsort(np.abs(vals))
    vals = vals[ind]
    vects = vects[:, ind]
    vects = vects[matrix_size:, :]

    # vals, vects = surface_greens_function_poles(E, h_l, h_0, h_r, damp)
    num_eigs = len(vals)

    u_right = vects[:, :num_eigs // 2]
    u_left = vects[:, num_eigs - 1:num_eigs // 2 - 1:-1]
    lambda_right = np.diag(vals[:num_eigs // 2])
    lambda_left = np.diag(vals[num_eigs - 1:num_eigs // 2 - 1:-1])

    h0 = -main_matrix[matrix_size:2*matrix_size, matrix_size:2*matrix_size] + (E + damp) * np.identity(matrix_size)
    hl = -main_matrix[matrix_size:2*matrix_size, 0:matrix_size]
    hr = overlap_matrix[matrix_size:2*matrix_size, matrix_size:2*matrix_size]

    sgf_l = hr.dot(u_right).dot(lambda_right).dot(np.linalg.pinv(u_right))
    sgf_r = hl.dot(u_left).dot(np.linalg.inv(lambda_left)).dot(np.linalg.pinv(u_left))

    if iterate:
        sgf_l = iterate_gf(E, h0, hl, hr, sgf_l, 2)
        sgf_r = iterate_gf(E, h0, hr, hl, sgf_r, 2)

    s01, s02 = h_0[0].shape
    s11, s12 = h_0[-1].shape

    sgf_l = sgf_l[-s11:, -s12:]
    sgf_r = sgf_r[:s01, :s02]

    return sgf_r, sgf_l


def simple_iterative_greens_function(E, h_l, h_0, h_r, **kwargs):
    """Computes the self-energy matrix using the canonical mixing technique.
    Parameters
    ----------
    E : float
        energy

    h_l : numpy array
          left block of three-block-diagonal Hamiltonian

    h_0 : numpy array
          central block of three-block-diagonal Hamiltonian

    h_r : numpy array
          right block of three-block-diagonal Hamiltonian

    kwargs : dict
          additional named arguments:
            s_l : numpy array (Default value = zero matrix)
                  left block of three-block-diagonal overlap matrix

            s_0 : numpy array (Default value = identity matrix)
                  central block of three-block-diagonal overlap matrix

            s_r : numpy array (Default value = zero matrix)
                  right block of three-block-diagonal overlap matrix

          alpha : float (Default value = 0.5)
                  mixing parameter (0 < alpha < 1) for green's function

           damp : float (Default value = 1e-6)
                  positive infinitesimal a.k.a broadening parameter
                  that shifts solution off real axis

   initialguess : numpy array (Default value = zero matrix)
                  initial guess of the self-energy, allows restarting of algorithm
                  to gain increased precision, or if the user has a good idea of
                  what the self-energy should be, can boost convergence

          nconv : float (Default value = 1e-10)
                  convergence requirement for iteration using the 2-norm,
                  also known as the largest singular value.

    Returns
    -------
    numpy.matrix

        #Guide for usage.

        A user provides the energy; onsite-, hopping-, and overlap- matrices;
        the mixing parameter; the broadening parameter; initial guess (optional);
        and convergence criterion in order to compute the retarded self-energy.

        Note that due to its singular nature, computing the self-energy of a
        matrix with damp=0 is numerically undefined when starting from scratch,
        however, if the user desires, they may compute the solution at some small
        damp, say 1e-8, then use that as an input guess for a damp=0 input,
        the imaginary part of the initial guess shifts the solution off the real
        line and allows for convergence. We do not suggest starting a damp=0
        iteration from an arbitrary guess due to instability in convergence.

    """

    s_l = kwargs.get('s_l', np.zeros(h_0.shape[0]))
    s_0 = kwargs.get('s_0', np.identity(h_0.shape[0]))
    s_r = kwargs.get('s_r', np.zeros(h_0.shape[0]))
    alpha = kwargs.get('alpha', 0.5)
    damp = kwargs.get('damp', 1e-6)
    initialguess = kwargs.get('initialguess', np.zeros(h_0.shape[0]))
    nconv = kwargs.get('nconv', 1e-10)

    # Check inputs:
    # If alpha is in an unphysical range, fix it to 0.5:
    if (alpha < 0) or (alpha > 1):
        alpha = 0.5
    else:
        alpha = alpha

    # If damp is negative or imaginary set it to its modulus
    damp = np.abs(damp)

    # Define initial K-matrices
    KL = (E + 1j * damp) * s_l - h_l
    K0 = (E + 1j * damp) * s_0 - h_0
    KR = (E + 1j * damp) * s_r - h_r

    # Form initial guess
    AN = K0 - initialguess

    # initialize iteration parameters
    NewSelfEnergy = initialguess  #
    convcheck = 1  # dummy parameter much larger than nconv
    ctN = 0  # parameter to count how many iterations it took

    while convcheck > nconv:
        ctN = ctN + 1

        # To compare between iterations
        OldSelfEnergy = NewSelfEnergy
        # Factorize and solve inverse
        lu, piv = linalg.lu_factor(AN)
        XN = linalg.lu_solve((lu, piv), KR)
        NewSelfEnergy = KL.dot(XN)
        # Check magnitude of change, if too big, mix in new solution
        convcheck = linalg.norm(OldSelfEnergy - NewSelfEnergy, 2)
        AN = (1 - alpha) * AN + alpha * (K0 - KL.dot(XN))
        # print(convcheck) #for debugging

    # other outputs could the green's function if a flag is set because it requires
    # another inverse or the value of convcheck

    return NewSelfEnergy


def sancho_rubio_iterative_greens_function(E, h_l, h_0, h_r, **kwargs):
    """Computes the self-energy matrix using the Sancho-Rubio technique.
    Parameters
    ----------
    E : float
        energy

    h_l : numpy array
          left block of three-block-diagonal Hamiltonian

    h_0 : numpy array
          central block of three-block-diagonal Hamiltonian

    h_r : numpy array
          right block of three-block-diagonal Hamiltonian

    kwargs : dict
          additional named arguments:
            s_l : numpy array (Default value = zero matrix)
                  left block of three-block-diagonal overlap matrix

            s_0 : numpy array (Default value = identity matrix)
                  central block of three-block-diagonal overlap matrix

            s_r : numpy array (Default value = zero matrix)
                  right block of three-block-diagonal overlap matrix
                  
           damp : float (Default value = 1e-6)
                  positive infinitesimal a.k.a broadening parameter
                  that shifts solution off real axis

          nconv : float (Default value = 1e-10)
                  convergence requirement for iteration using the 2-norm,
                  also known as the largest singular value.

    Returns
    -------
    numpy.matrix

        #Guide for usage.

        A user provides the energy; onsite-, hopping-, and overlap- 
        matrices; the broadening parameter; and convergence criterion
        in order to compute the retarded self-energy.

    """

    s_l = kwargs.get('s_l', np.zeros(h_0.shape[0]))
    s_0 = kwargs.get('s_0', np.identity(h_0.shape[0]))
    s_r = kwargs.get('s_r', np.zeros(h_0.shape[0]))
    damp = kwargs.get('damp', 1e-6)
    nconv = kwargs.get('nconv', 1e-10)

    # If damp is negative or imaginary set it to its modulus
    damp = np.abs(damp)

    # Define initial K-matrices
    KL = (E + 1j * damp) * s_l - h_l
    K0 = (E + 1j * damp) * s_0 - h_0
    KR = (E + 1j * damp) * s_r - h_r

    #Seed values
    alpha0 = KR
    beta0 = KL
    esOld = K0
    eOld = esOld
    alpha = alpha0
    beta  = beta0

    itr = 0        # initialise iteration check
    normcheck = 1  # initialise normalisation check
    maxiter=100

    while (itr < maxiter) & (normcheck > nconv):
        itr = itr + 1
        
        lu, piv = linalg.lu_factor(eOld)
        a = linalg.lu_solve((lu, piv), alpha)
        b = linalg.lu_solve((lu, piv), beta)
        
        alphab = alpha.dot(b)
        esNew = esOld - alphab
        eNew = eOld - beta.dot(a) - alphab
        alpha = alpha.dot(a)
        beta = beta.dot(b)
        normcheck = linalg.norm(esNew-esOld, 2)
        eOld  = eNew
        esOld = esNew
        # print(normcheck) #for debugging
 
    
    lu, piv = linalg.lu_factor(esNew)
    SelfEnergy = alpha0.dot( linalg.lu_solve( (lu, piv), beta0) )

    return SelfEnergy


import logging
import warnings


VERBOSITY = 1


def set_verbosity(level):
    global VERBOSITY
    if level > 0:
        logging.getLogger().setLevel(logging.INFO)
        VERBOSITY = level
    else:
        warnings.filterwarnings("ignore")
        logging.getLogger().setLevel(logging.CRITICAL)
        VERBOSITY = level



import numpy as np


def fermi_window(energy, tempr):
    """

    Parameters
    ----------
    energy :
        
    tempr :
        

    Returns
    -------

    """

    kb = 8.61733e-5  # Boltzmann constant in eV

    return (1.0 / (4.0 * kb * tempr)) / np.cosh(energy / (2 * kb * tempr))


def tr2cond(energy, tr, tempr=300):
    """

    Parameters
    ----------
    energy :
        
    tr :
        
    tempr :
         (Default value = 300)

    Returns
    -------

    """

    tr = np.pad(tr, 30000, 'edge')
    energy = np.pad(energy, 30000,'linear_ramp')
    ans = np.convolve(tr, fermi_window(energy, tempr), mode='same')

    return energy, ans